#! /usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of IVRE.
# Copyright 2011 - 2018 Pierre LALET <pierre.lalet@cea.fr>
#
# IVRE is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IVRE is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with IVRE. If not, see <http://www.gnu.org/licenses/>.

"""This sub-module contains functions to interact with PostgreSQL
databases.

"""


import datetime
import time


from sqlalchemy import ARRAY, Column, Index, LargeBinary, String, Table, \
    and_, cast, column, delete, desc, exists, func, insert, join, not_, \
    nullsfirst, select, text, tuple_, update
from sqlalchemy.dialects import postgresql


from ivre import config, utils, xmlnmap
from ivre.db.sql import PassiveCSVFile, ScanCSVFile, SQLDB, SQLDBActive, \
    SQLDBFlow, SQLDBNmap, SQLDBPassive, SQLDBView


class PostgresDB(SQLDB):

    def __init__(self, url):
        super(PostgresDB, self).__init__(url)

    @staticmethod
    def ip2internal(addr):
        return utils.force_int2ip(addr)

    @staticmethod
    def internal2ip(addr):
        return addr

    def copy_from(self, *args, **kargs):
        cursor = self.db.raw_connection().cursor()
        conn = self.db.connect()
        trans = conn.begin()
        cursor.copy_from(*args, **kargs)
        trans.commit()
        conn.close()

    def create_tmp_table(self, table, extracols=None):
        cols = [c.copy() for c in table.__table__.columns]
        for c in cols:
            c.index = False
            c.nullable = True
            c.foreign_keys = None
            if c.primary_key:
                c.primary_key = False
                c.index = True
        if extracols is not None:
            cols.extend(extracols)
        t = Table("tmp_%s" % table.__tablename__,
                  table.__table__.metadata, *cols,
                  prefixes=['TEMPORARY'])
        t.create(bind=self.db, checkfirst=True)
        return t

    def start_bulk_insert(self, size=None, retries=0):
        return BulkInsert(self.db, size=size, retries=retries)

    def explain(self, req, **_):
        req_comp = req.compile(dialect=postgresql.dialect())
        arg_dic = {}
        for k in req_comp.params:
            arg_dic[k] = repr(req_comp.params[k])
        req_cur = self.db.execute(text("EXPLAIN " + req_comp.string % arg_dic))
        return "\n".join(map(" ".join, req_cur.fetchall()))


class BulkInsert(object):
    """A PostgreSQL transaction, with automatic commits"""

    def __init__(self, db, size=None, retries=0):
        """`size` is the number of inserts per commit and `retries` is the
        number of times to retry a failed transaction (when inserting
        concurrently for example). 0 is forever, 1 does not retry, 2 retries
        once, etc.
        """
        self.db = db
        self.start_time = time.time()
        self.commited_counts = {}
        self.size = config.POSTGRES_BATCH_SIZE if size is None else size
        self.retries = retries
        self.conn = db.connect()
        self.trans = self.conn.begin()
        self.queries = {}

    def append(self, query):
        query._set_bind(self.db)
        s_query = str(query)
        params = query.parameters
        query.parameters = None
        self.queries.setdefault(s_query,
                                (query, []))[1].append(params)
        if len(self.queries[s_query][1]) >= self.size:
            self.commit(query=s_query)

    def commit(self, query=None, renew=True):
        if query is None:
            last = len(self.queries) - 1
            for i, query in enumerate(list(self.queries)):
                self.commit(query=query, renew=True if i < last else renew)
            return
        q_query, params = self.queries.pop(query)
        self.conn.execute(q_query, *params)
        self.trans.commit()
        newtime = time.time()
        l_params = len(params)
        try:
            self.commited_counts[query] += l_params
        except KeyError:
            self.commited_counts[query] = l_params
        rate = l_params / (newtime - self.start_time)
        utils.LOGGER.debug("DB:%s", query)
        utils.LOGGER.debug("DB:%d inserts, %f/sec (total %d)",
                           l_params, rate, self.commited_counts[query])
        if renew:
            self.start_time = newtime
            self.trans = self.conn.begin()

    def close(self):
        self.commit(renew=False)
        self.conn.close()


class PostgresDBFlow(PostgresDB, SQLDBFlow):

    def __init__(self, url):
        super(PostgresDBFlow, self).__init__(url)


class PostgresDBActive(PostgresDB, SQLDBActive):

    def __init__(self, url):
        super(PostgresDBActive, self).__init__(url)

    def __migrate_schema_10_11(self):
        """Converts a record from version 10 to version 11.

The PostgreSQL backend is only conerned by a limited subset of
changes.

        """
        req = (select([self.tables.scan.id,
                       self.tables.script.port,
                       self.tables.script.output,
                       self.tables.script.data])
               .select_from(join(join(self.tables.scan, self.tables.port),
                                 self.tables.script))
               .where(and_(
                   self.tables.scan.schema_version == 10,
                   self.tables.script.name == "ssl-cert",
               )))
        for rec in self.db.execute(req):
            if 'ssl-cert' in rec.data:
                if 'pem' in rec.data['ssl-cert']:
                    data = ''.join(
                        rec.data['ssl-cert']['pem'].splitlines()[1:-1]
                    ).encode()
                    try:
                        newout, newinfo = xmlnmap.create_ssl_cert(data)
                    except Exception:
                        utils.LOGGER.warning('Cannot parse certificate %r',
                                             data,
                                             exc_info=True)
                    else:
                        self.db.execute(
                            update(self.tables.script)
                            .where(and_(self.tables.script.port == rec.port,
                                        self.tables.script.name == "ssl-cert"))
                            .values(data={"ssl-cert": newinfo},
                                    output='\n'.join(newout))
                        )
                        continue
                    try:
                        pubkeytype = {
                            'rsaEncryption': 'rsa',
                            'id-ecPublicKey': 'ec',
                            'id-dsa': 'dsa',
                            'dhpublicnumber': 'dh',
                        }[rec.data['ssl-cert'].pop('pubkeyalgo')]
                    except KeyError:
                        pass
                    else:
                        self.db.execute(
                            update(self.tables.script)
                            .where(and_(self.tables.script.port == rec.port,
                                        self.tables.script.name == "ssl-cert"))
                            .values(data={"ssl-cert":
                                          dict(rec.data['ssl-cert'],
                                               type=pubkeytype)})
                        )
        self.db.execute(
            update(self.tables.scan)
            .where(self.tables.scan.schema_version == 10)
            .values(schema_version=11)
        )
        return 0

    def start_store_hosts(self):
        """Backend-specific subclasses may use this method to create some bulk
insert structures.

        """
        self.bulk = self.start_bulk_insert()

    def stop_store_hosts(self):
        """Backend-specific subclasses may use this method to commit bulk
insert structures.

        """
        self.bulk.close()
        self.bulk = None

    def _get_ips_ports(self, flt, limit=None, skip=None):
        req = flt.query(select([self.tables.scan.id]))
        if skip is not None:
            req = req.offset(skip)
        if limit is not None:
            req = req.limit(limit)
        base = req.cte("base")
        return (
            {
                "addr": rec[2], "starttime": rec[1],
                "ports": [
                    {"proto": proto, "port": int(port), "state_state": state}
                    for proto, port, state in (
                        elt.split(',') for elt in
                        ''.join(rec[0])[3:-3].split(')","(')
                    )
                ]
            }
            for rec in
            self.db.execute(
                select([
                    func.array_agg(postgresql.aggregate_order_by(
                        tuple_(self.tables.port.protocol,
                               self.tables.port.port,
                               self.tables.port.state).label('a'),
                        tuple_(self.tables.port.protocol,
                               self.tables.port.port).label('a')
                    )).label('ports'),
                    self.tables.scan.time_start, self.tables.scan.addr,
                ])
                .select_from(join(self.tables.port, self.tables.scan))
                .group_by(self.tables.scan.addr, self.tables.scan.time_start)
                .where(and_(self.tables.port.port >= 0,
                            self.tables.scan.id.in_(base)))
            )
        )

    def get_ips_ports(self, flt, limit=None, skip=None):
        result = list(self._get_ips_ports(flt, limit=limit, skip=skip))
        return result, sum(len(host.get('ports', [])) for host in result)

    def topvalues(self, field, flt=None, topnbr=10, sort=None,
                  limit=None, skip=None, least=False):
        """
        This method makes use of the aggregation framework to produce
        top values for a given field or pseudo-field. Pseudo-fields are:
          - category / label / asnum / country / net[:mask]
          - port
          - port:open / :closed / :filtered / :<servicename>
          - portlist:open / :closed / :filtered
          - countports:open / :closed / :filtered
          - service / service:<portnbr>
          - product / product:<portnbr>
          - cpe / cpe.<part> / cpe:<cpe_spec> / cpe.<part>:<cpe_spec>
          - devicetype / devicetype:<portnbr>
          - script:<scriptid> / script:<port>:<scriptid>
            / script:host:<scriptid>
          - cert.* / smb.* / sshkey.*
          - httphdr / httphdr.{name,value} / httphdr:<name>
          - modbus.* / s7.* / enip.*
          - mongo.dbs.*
          - vulns.*
          - screenwords
          - file.* / file.*:scriptid
          - hop
        """
        if flt is None:
            flt = self.flt_empty
        base = flt.query(
            select([self.tables.scan.id]).select_from(flt.select_from)
        ).cte("base")
        order = "count" if least else desc("count")
        outputproc = None
        if field == "port":
            field = self._topstructure(self.tables.port,
                                       [self.tables.port.protocol,
                                        self.tables.port.port],
                                       self.tables.port.state == "open")
        elif field == "ttl":
            field = self._topstructure(
                self.tables.port, [self.tables.port.state_reason_ttl],
                self.tables.port.state_reason_ttl != None,
                # noqa: E711 (BinaryExpression)
            )
        elif field == "ttlinit":
            field = self._topstructure(
                self.tables.port,
                [func.least(255, func.power(2, func.ceil(
                    func.log(2, self.tables.port.state_reason_ttl)
                )))],
                self.tables.port.state_reason_ttl != None,
                # noqa: E711 (BinaryExpression)
            )
            outputproc = int
        elif field.startswith('port:'):
            info = field[5:]
            field = self._topstructure(
                self.tables.port,
                [self.tables.port.protocol, self.tables.port.port],
                (self.tables.port.state == info)
                if info in ['open', 'filtered', 'closed', 'open|filtered'] else
                (self.tables.port.service_name == info),
            )
        elif field.startswith('countports:'):
            info = field[11:]
            return (
                {"count": result[0], "_id": result[1]}
                for result in self.db.execute(
                    select([func.count().label("count"),
                            column('cnt')])
                    .select_from(
                        select([func.count().label('cnt')])
                        .select_from(self.tables.port)
                        .where(and_(
                            self.tables.port.state == info,
                            # self.tables.port.scan.in_(base),
                            exists(
                                select([1])\
                                .select_from(base)\
                                .where(
                                    self.tables.port.scan == base.c.id
                                )
                            ),
                        ))\
                        .group_by(self.tables.port.scan)\
                        .alias('cnt')
                    ).group_by('cnt').order_by(order).limit(topnbr)
                )
            )
        elif field.startswith('portlist:'):
            # Deux options pour filtrer:
            #   -1- self.tables.port.scan.in_(base),
            #   -2- exists(select([1])\
            #       .select_from(base)\
            #       .where(
            #         self.tables.port.scan == base.c.id
            #       )),
            #
            # D'après quelques tests, l'option -1- est plus beaucoup
            # rapide quand (base) est pas ou peu sélectif, l'option
            # -2- un peu plus rapide quand (base) est très sélectif
            #
            # TODO: vérifier si c'est pareil pour:
            #  - countports:open
            #  - tous les autres
            info = field[9:]
            return (
                {"count": result[0], "_id": [
                    (proto, int(port)) for proto, port in (
                        elt.split(',') for elt in
                        result[1][3:-3].split(')","(')
                    )
                ]}
                for result in self.db.execute(
                    select([func.count().label("count"),
                            column('ports')])
                    .select_from(
                        select([
                            func.array_agg(postgresql.aggregate_order_by(
                                tuple_(self.tables.port.protocol,
                                       self.tables.port.port).label('a'),
                                tuple_(self.tables.port.protocol,
                                       self.tables.port.port).label('a')
                            )).label('ports'),
                        ])
                        .where(and_(
                            self.tables.port.state == info,
                            self.tables.port.scan.in_(base),
                            # exists(select([1])\
                            #        .select_from(base)\
                            #        .where(
                            #            self.tables.port.scan == base.c.id
                            #        )),
                        ))
                        .group_by(self.tables.port.scan)
                        .alias('ports')
                    )
                    .group_by('ports').order_by(order).limit(topnbr)
                )
            )
        elif field == "service":
            field = self._topstructure(self.tables.port,
                                       [self.tables.port.service_name],
                                       self.tables.port.state == "open")
        elif field.startswith("service:"):
            info = field[8:]
            if '/' in info:
                info = info.split('/', 1)
                field = self._topstructure(
                    self.tables.port, [self.tables.port.service_name],
                    and_(self.tables.port.protocol == info[0],
                         self.tables.port.port == int(info[1])),
                )
            else:
                field = self._topstructure(self.tables.port,
                                           [self.tables.port.service_name],
                                           self.tables.port.port == int(info))
        elif field == "product":
            field = self._topstructure(
                self.tables.port,
                [self.tables.port.service_name,
                 self.tables.port.service_product],
                self.tables.port.state == "open",
            )
        elif field.startswith("product:"):
            info = field[8:]
            if info.isdigit():
                info = int(info)
                flt = self.flt_and(flt, self.searchport(info))
                field = self._topstructure(
                    self.tables.port,
                    [self.tables.port.service_name,
                     self.tables.port.service_product],
                    and_(self.tables.port.state == "open",
                         self.tables.port.port == info),
                )
            elif info.startswith('tcp/') or info.startswith('udp/'):
                info = (info[:3], int(info[4:]))
                flt = self.flt_and(flt, self.searchport(info[1],
                                                        protocol=info[0]))
                field = self._topstructure(
                    self.tables.port,
                    [self.tables.port.service_name,
                     self.tables.port.service_product],
                    and_(self.tables.port.state == "open",
                         self.tables.port.port == info[1],
                         self.tables.port.protocol == info[0]),
                )
            else:
                flt = self.flt_and(flt, self.searchservice(info))
                field = self._topstructure(
                    self.tables.port,
                    [self.tables.port.service_name,
                     self.tables.port.service_product],
                    and_(self.tables.port.state == "open",
                         self.tables.port.service_name == info),
                )
        elif field == "devicetype":
            field = self._topstructure(self.tables.port,
                                       [self.tables.port.service_devicetype],
                                       self.tables.port.state == "open")
        elif field.startswith("devicetype:"):
            info = field[11:]
            if info.isdigit():
                info = int(info)
                flt = self.flt_and(flt, self.searchport(info))
                field = self._topstructure(
                    self.tables.port,
                    [self.tables.port.service_devicetype],
                    and_(self.tables.port.state == "open",
                         self.tables.port.port == info)
                )
            elif info.startswith('tcp/') or info.startswith('udp/'):
                info = (info[:3], int(info[4:]))
                flt = self.flt_and(flt, self.searchport(info[1],
                                                        protocol=info[0]))
                field = self._topstructure(
                    self.tables.port,
                    [self.tables.port.service_devicetype],
                    and_(self.tables.port.state == "open",
                         self.tables.port.port == info[1],
                         self.tables.port.protocol == info[0])
                )
            else:
                flt = self.flt_and(flt, self.searchservice(info))
                field = self._topstructure(
                    self.tables.port,
                    [self.tables.port.service_devicetype],
                    and_(self.tables.port.state == "open",
                         self.tables.port.service_name == info)
                )
        elif field == "version":
            field = self._topstructure(
                self.tables.port,
                [self.tables.port.service_name,
                 self.tables.port.service_product,
                 self.tables.port.service_version],
                self.tables.port.state == "open",
            )
        elif field.startswith("version:"):
            info = field[8:]
            if info.isdigit():
                info = int(info)
                flt = self.flt_and(flt, self.searchport(info))
                field = self._topstructure(
                    self.tables.port,
                    [self.tables.port.service_name,
                     self.tables.port.service_product,
                     self.tables.port.service_version],
                    and_(self.tables.port.state == "open",
                         self.tables.port.port == info),
                )
            elif info.startswith('tcp/') or info.startswith('udp/'):
                info = (info[:3], int(info[4:]))
                flt = self.flt_and(flt, self.searchport(info[1],
                                                        protocol=info[0]))
                field = self._topstructure(
                    self.tables.port,
                    [self.tables.port.service_name,
                     self.tables.port.service_product,
                     self.tables.port.service_version],
                    and_(self.tables.port.state == "open",
                         self.tables.port.port == info[1],
                         self.tables.port.protocol == info[0]),
                )
            elif ':' in info:
                info = info.split(':', 1)
                flt = self.flt_and(flt, self.searchproduct(info[1],
                                                           service=info[0]))
                field = self._topstructure(
                    self.tables.port,
                    [self.tables.port.service_name,
                     self.tables.port.service_product,
                     self.tables.port.service_version],
                    and_(self.tables.port.state == "open",
                         self.tables.port.service_name == info[0],
                         self.tables.port.service_product == info[1]),
                )
            else:
                flt = self.flt_and(flt, self.searchservice(info))
                field = self._topstructure(
                    self.tables.port,
                    [self.tables.port.service_name,
                     self.tables.port.service_product,
                     self.tables.port.service_version],
                    and_(self.tables.port.state == "open",
                         self.tables.port.service_name == info),
                )
        elif field == "asnum":
            field = self._topstructure(self.tables.scan,
                                       [self.tables.scan.info["as_num"]])
        elif field == "as":
            field = self._topstructure(self.tables.scan,
                                       [self.tables.scan.info["as_num"],
                                        self.tables.scan.info["as_name"]])
        elif field == "country":
            field = self._topstructure(self.tables.scan,
                                       [self.tables.scan.info["country_code"],
                                        self.tables.scan.info["country_name"]])
        elif field == "city":
            field = self._topstructure(self.tables.scan,
                                       [self.tables.scan.info["country_code"],
                                        self.tables.scan.info["city"]])
        elif field == "net" or field.startswith("net:"):
            info = field[4:]
            info = int(info) if info else 24
            field = self._topstructure(
                self.tables.scan,
                [func.set_masklen(text("scan.addr::cidr"), info)],
            )
        elif field == "script" or field.startswith("script:"):
            info = field[7:]
            if info:
                field = self._topstructure(self.tables.script,
                                           [self.tables.script.output],
                                           self.tables.script.name == info)
            else:
                field = self._topstructure(self.tables.script,
                                           [self.tables.script.name])
        elif field in ["category", "categories"]:
            field = self._topstructure(self.tables.category,
                                       [self.tables.category.name])
        elif field.startswith('cert.'):
            subfield = field[5:]
            field = self._topstructure(
                self.tables.script,
                [self.tables.script.data['ssl-cert'][subfield]],
                and_(self.tables.script.name == 'ssl-cert',
                     self.tables.script.data['ssl-cert'].has_key(subfield))
            )  # noqa: W601 (BinaryExpression)
        elif field == 'useragent' or field.startswith('useragent:'):
            if field == 'useragent':
                flt = self.flt_and(flt, self.searchuseragent())
                field = self._topstructure(
                    self.tables.script,
                    [column('http_user_agent')],
                    self.tables.script.name == 'http-user-agent',
                    None,
                    func.jsonb_array_elements(
                        self.tables.script.data['http-user-agent'],
                    ).alias('http_user_agent')
                )  # noqa: W601 (BinaryExpression)
            else:
                subfield = utils.str2regexp(field[10:])
                flt = self.flt_and(flt,
                                   self.searchuseragent(useragent=subfield))
                field = self._topstructure(
                    self.tables.script,
                    [column('http_user_agent')],
                    and_(self.tables.script.name == 'http-user-agent',
                         self._searchstring_re(
                             column('http_user_agent').op('->>')(0),
                             subfield,
                         )),
                    None,
                    func.jsonb_array_elements(
                        self.tables.script.data['http-user-agent'],
                    ).alias('http_user_agent')
                )  # noqa: W601 (BinaryExpression)
        elif field == 'ja3-client' or (
                field.startswith('ja3-client') and field[10] in ':.'
        ):
            if ':' in field:
                field, value = field.split(':', 1)
                subkey, value = self._ja3keyvalue(utils.str2regexp(value))
            else:
                value = None
            if '.' in field:
                field, subfield = field.split('.', 1)
            else:
                subfield = 'md5'
            flt = self.flt_and(flt, self.searchja3client(value_or_hash=value))
            if value is None:
                field = self._topstructure(
                    self.tables.script,
                    [column('ssl_ja3_client').op('->>')(subfield)],
                    self.tables.script.name == 'ssl-ja3-client',
                    None,
                    func.jsonb_array_elements(
                        self.tables.script.data['ssl-ja3-client'],
                    ).alias('ssl_ja3_client')
                )  # noqa: W601 (BinaryExpression)
            else:
                field = self._topstructure(
                    self.tables.script,
                    [column('ssl_ja3_client').op('->>')(subfield)],
                    and_(self.tables.script.name == 'ssl-ja3-client',
                         self._searchstring_re(
                             column('ssl_ja3_client').op('->>')(subkey),
                             value,
                         )),
                    None,
                    func.jsonb_array_elements(
                        self.tables.script.data['ssl-ja3-client'],
                    ).alias('ssl_ja3_client')
                )  # noqa: W601 (BinaryExpression)
        elif field == 'ja3-server' or (
                field.startswith('ja3-server') and field[10] in ':.'
        ):
            if ':' in field:
                field, values = field.split(':', 1)
                if ':' in values:
                    value1, value2 = values.split(':', 1)
                    if value1:
                        subkey1, value1 = self._ja3keyvalue(
                            utils.str2regexp(value1)
                        )
                    else:
                        subkey1, value1 = None, None
                    if value2:
                        subkey2, value2 = self._ja3keyvalue(
                            utils.str2regexp(value2)
                        )
                    else:
                        subkey2, value2 = None, None
                else:
                    subkey1, value1 = self._ja3keyvalue(
                        utils.str2regexp(values)
                    )
                    subkey2, value2 = None, None
            else:
                subkey1, value1 = None, None
                subkey2, value2 = None, None
            if '.' in field:
                field, subfield = field.split('.', 1)
            else:
                subfield = 'md5'
            condition = self.tables.script.name == 'ssl-ja3-server'
            if value1 is not None:
                condition = and_(
                    condition,
                    self._searchstring_re(
                        column('ssl_ja3_server').op('->>')(subkey1),
                        value1,
                    )
                )
            if value2 is not None:
                condition = and_(
                    condition,
                    self._searchstring_re(
                        column('ssl_ja3_server').op('->')('client')
                        .op('->>')(subkey2),
                        value2,
                    )
                )
            field = self._topstructure(
                self.tables.script,
                [column('ssl_ja3_server').op('->>')(subfield),
                 column('ssl_ja3_server').op('->')('client')
                 .op('->>')(subfield)],
                condition,
                None,
                func.jsonb_array_elements(
                    self.tables.script.data['ssl-ja3-server'],
                ).alias('ssl_ja3_server')
            )  # noqa: W601 (BinaryExpression)
        elif field == "source":
            field = self._topstructure(self.tables.scan,
                                       [self.tables.scan.source])
        elif field == "domains":
            field = self._topstructure(
                self.tables.hostname,
                [func.unnest(self.tables.hostname.domains)]
            )
        elif field.startswith("domains:"):
            level = int(field[8:]) - 1
            base1 = (select([func.unnest(self.tables.hostname.domains)
                             .label("domains")])
                     .where(exists(select([1])
                                   .select_from(base)
                                   .where(self.tables.hostname.scan ==
                                          base.c.id)))
                     .cte("base1"))
            return (
                {"count": result[1], "_id": result[0]}
                for result in self.db.execute(
                    select([base1.c.domains, func.count().label("count")])
                    .where(base1.c.domains.op('~')(
                        '^([^\\.]+\\.){%d}[^\\.]+$' % level
                    ))
                    .group_by(base1.c.domains)
                    .order_by(order)
                    .limit(topnbr)
                )
            )
        elif field == "hop":
            field = self._topstructure(self.tables.hop,
                                       [self.tables.hop.ipaddr])
        elif field.startswith('hop') and field[3] in ':>':
            ttl = int(field[4:])
            field = self._topstructure(
                self.tables.hop, [self.tables.hop.ipaddr],
                (self.tables.hop.ttl > ttl) if field[3] == '>' else
                (self.tables.hop.ttl == ttl),
            )
        elif field == 'file' or (field.startswith('file') and
                                 field[4] in '.:'):
            if field.startswith('file:'):
                scripts = field[5:]
                if '.' in scripts:
                    scripts, field = scripts.split('.', 1)
                else:
                    field = 'filename'
                scripts = scripts.split(',')
                flt = (self.tables.script.name == scripts[0]
                       if len(scripts) == 1 else
                       self.tables.script.name.in_(scripts))
            else:
                field = field[5:] or 'filename'
                flt = True
            field = self._topstructure(
                self.tables.script,
                [func.jsonb_array_elements(
                    func.jsonb_array_elements(
                        self.tables.script.data['ls']['volumes']
                    ).op('->')('files')
                ).op('->>')(field).label(field)],
                and_(
                    flt,
                    self.tables.script.data.op('@>')(
                        '{"ls": {"volumes": [{"files": []}]}}'
                    ),
                ),
            )
        elif field.startswith('modbus.'):
            subfield = field[7:]
            field = self._topstructure(
                self.tables.script,
                [self.tables.script.data['modbus-discover'][subfield]],
                and_(self.tables.script.name == 'modbus-discover',
                     self.tables.script.data['modbus-discover']
                     .has_key(subfield)),
                # noqa: W601 (BinaryExpression)
            )
        elif field.startswith('s7.'):
            subfield = field[3:]
            field = self._topstructure(
                self.tables.script,
                [self.tables.script.data['s7-info'][subfield]],
                and_(self.tables.script.name == 's7-info',
                     self.tables.script.data['s7-info'].has_key(subfield)),
                # noqa: W601 (BinaryExpression)
            )
        elif field == 'httphdr':
            flt = self.flt_and(flt, self.searchscript(name="http-headers"))
            field = self._topstructure(
                self.tables.script,
                [column("hdr").op('->>')('name').label("name"),
                 column("hdr").op('->>')('value').label("value")],
                self.tables.script.name == 'http-headers',
                [column("name"), column("value")],
                func.jsonb_array_elements(
                    self.tables.script.data['http-headers']
                ).alias('hdr'),
            )
        elif field.startswith('httphdr.'):
            flt = self.flt_and(flt, self.searchscript(name="http-headers"))
            field = self._topstructure(
                self.tables.script,
                [column("hdr").op('->>')(field[8:]).label("topvalue")],
                self.tables.script.name == 'http-headers',
                [column("topvalue")],
                func.jsonb_array_elements(
                    self.tables.script.data['http-headers']
                ).alias('hdr'),
            )
        elif field.startswith('httphdr:'):
            flt = self.flt_and(flt, self.searchhttphdr(name=field[8:].lower()))
            field = self._topstructure(
                self.tables.script,
                [column("hdr").op('->>')("value").label("value")],
                and_(self.tables.script.name == 'http-headers',
                     column("hdr").op('->>')("name") == field[8:].lower()),
                [column("value")],
                func.jsonb_array_elements(
                    self.tables.script.data['http-headers']
                ).alias('hdr'),
            )
        else:
            raise NotImplementedError()
        s_from = {
            self.tables.script: join(self.tables.script, self.tables.port),
            self.tables.port: self.tables.port,
            self.tables.category: join(self.tables.association_scan_category,
                                       self.tables.category),
            self.tables.hostname: self.tables.hostname,
            self.tables.hop: join(self.tables.trace, self.tables.hop),
        }
        where_clause = {
            self.tables.script: self.tables.port.scan == base.c.id,
            self.tables.port: self.tables.port.scan == base.c.id,
            self.tables.category:
            self.tables.association_scan_category.scan == base.c.id,
            self.tables.hostname: self.tables.hostname.scan == base.c.id,
            self.tables.hop: self.tables.trace.scan == base.c.id
        }
        if field.base == self.tables.scan:
            req = flt.query(
                select([func.count().label("count")] + field.fields)
                .select_from(self.tables.scan)
                .group_by(*field.fields)
            )
        else:
            req = (select([func.count().label("count")] + field.fields)
                   .select_from(s_from[field.base]))
            if field.extraselectfrom is not None:
                req = req.select_from(field.extraselectfrom)
            req = (req
                   .group_by(*(field.fields if field.group_by is None
                               else field.group_by))
                   .where(exists(select([1]).select_from(base)
                                 .where(where_clause[field.base]))))
        if field.where is not None:
            req = req.where(field.where)
        if outputproc is None:
            return ({"count": result[0],
                     "_id": result[1:] if len(result) > 2 else result[1]}
                    for result in
                    self.db.execute(req.order_by(order).limit(topnbr)))
        else:
            return ({"count": result[0],
                     "_id": outputproc(result[1:] if len(result) > 2
                                       else result[1])}
                    for result in
                    self.db.execute(req.order_by(order).limit(topnbr)))

    def _features_port_list(self, flt, yieldall, use_service,
                            use_product, use_version):
        base = flt.query(
            select([self.tables.scan.id]).select_from(flt.select_from)
        ).cte("base")
        if use_version:
            fields = [self.tables.port.port,
                      self.tables.port.service_name,
                      self.tables.port.service_product,
                      self.tables.port.service_version]
        elif use_product:
            fields = [self.tables.port.port,
                      self.tables.port.service_name,
                      self.tables.port.service_product]
        elif use_service:
            fields = [self.tables.port.port,
                      self.tables.port.service_name]
        else:
            fields = [self.tables.port.port]
        req = (
            select(fields)
            .group_by(*fields)
            .where(and_(
                exists(select([1]).select_from(base)
                       .where(self.tables.port.scan == base.c.id)),
                self.tables.port.state == "open",
                self.tables.port.port != -1,
            ))
        )
        if not yieldall:
            req = req.order_by(*(nullsfirst(fld) for fld in fields))
            return self.db.execute(req)
        else:
            # results will be modified, we cannot keep a RowProxy
            # instance, so we convert the results to lists
            return (list(rec) for rec in self.db.execute(req))

    def _features_port_get(self, features, flt, yieldall, use_service,
                           use_product, use_version):
        base = flt.query(
            select([self.tables.scan.id]).select_from(flt.select_from)
        ).cte("base")
        if use_version:
            fields = [cast(self.tables.port.port, String),
                      self.tables.port.service_name,
                      self.tables.port.service_product,
                      self.tables.port.service_version]
        elif use_product:
            fields = [cast(self.tables.port.port, String),
                      self.tables.port.service_name,
                      self.tables.port.service_product]
        elif use_service:
            fields = [cast(self.tables.port.port, String),
                      self.tables.port.service_name]
        else:
            fields = [self.tables.port.port]
        n_features = len(features)
        for addr, cur_features in self.db.execute(
                select([
                    self.tables.scan.addr,
                    func.array_agg(func.distinct(postgresql.array(fields))),
                ])
                .select_from(join(self.tables.scan, self.tables.port))
                .group_by(self.tables.scan.addr)
                .where(and_(
                    exists(select([1]).select_from(base)
                           .where(self.tables.port.scan == base.c.id)),
                    self.tables.port.state == "open",
                    self.tables.port.port != -1,
                ))
        ):
            currec = [0] * n_features
            for feat in cur_features:
                if use_service:
                    # convert port number back to an integer
                    feat[0] = int(feat[0])
                try:
                    currec[features[tuple(feat)]] = 1
                except KeyError:
                    pass
            yield (addr, currec)
        # add features for addresses without open ports
        base2 = flt.query(
            select([func.distinct(self.tables.port.scan).label('scan')])
            .select_from(flt.select_from)
            .where(and_(
                exists(select([1]).select_from(base)
                       .where(self.tables.port.scan == base.c.id)),
                self.tables.port.state == "open",
                self.tables.port.port != -1,
            ))
        ).cte("base2")
        for addr, in self.db.execute(
                flt.query(
                    select([func.distinct(self.tables.scan.addr)])
                    .select_from(flt.select_from)
                    .where(not_(exists(
                        select([1]).select_from(base2)
                        .where(self.tables.scan.id == base2.c.scan)
                    )))
                )
        ):
            print("ADDING RECORD FOR %r" % addr)
            yield (addr, [0] * n_features)


class PostgresDBNmap(PostgresDBActive, SQLDBNmap):

    def __init__(self, url):
        super(PostgresDBNmap, self).__init__(url)

    def store_scan_doc(self, scan):
        scan = scan.copy()
        if 'start' in scan:
            scan['start'] = datetime.datetime.utcfromtimestamp(
                int(scan['start'])
            )
        if 'scaninfos' in scan:
            scan["scaninfo"] = scan.pop('scaninfos')
        scan["sha256"] = utils.decode_hex(scan.pop('_id'))
        insrt = insert(self.tables.scanfile).values(
            **dict(
                (key, scan[key])
                for key in ['sha256', 'args', 'scaninfo', 'scanner', 'start',
                            'version', 'xmloutputversion']
                if key in scan
            )
        )
        if config.DEBUG:
            scanfileid = self.db.execute(
                insrt.returning(self.tables.scanfile.sha256)
            ).fetchone()[0]
            utils.LOGGER.debug("SCAN STORED: %r", utils.encode_hex(scanfileid))
        else:
            self.db.execute(insrt)

    def _store_host(self, host):
        addr = self.ip2internal(host['addr'])
        info = host.get('infos')
        source = host.get('source', '')
        host_tstart = utils.all2datetime(host['starttime'])
        host_tstop = utils.all2datetime(host['endtime'])
        scanid = self.db.execute(
            postgresql.insert(self.tables.scan).values(
                addr=addr,
                source=source,
                info=info,
                time_start=host_tstart,
                time_stop=host_tstop,
                # FIXME: masscan results may lack 'state' and 'state_reason'
                state=host.get('state'),
                state_reason=host.get('state_reason'),
                state_reason_ttl=host.get('state_reason_ttl'),
            )
            .on_conflict_do_nothing()
            .returning(self.tables.scan.id)
        ).fetchone()[0]
        for category in host.get("categories", []):
            insrt = postgresql.insert(self.tables.category)
            catid = self.db.execute(
                insrt.values(name=category)
                .on_conflict_do_update(
                    index_elements=['name'],
                    set_={'name': insrt.excluded.name}
                )
                .returning(self.tables.category.id)
            ).fetchone()[0]
            self.db.execute(postgresql.insert(
                self.tables.association_scan_category
            ).values(scan=scanid, category=catid)
             .on_conflict_do_nothing())
        for port in host.get('ports', []):
            scripts = port.pop('scripts', [])
            # FIXME: handle screenshots
            for fld in ['screendata', 'screenshot', 'screenwords',
                        'service_method']:
                try:
                    del port[fld]
                except KeyError:
                    pass
            if 'service_servicefp' in port:
                port['service_fp'] = port.pop('service_servicefp')
            if 'state_state' in port:
                port['state'] = port.pop('state_state')
            if 'state_reason_ip' in port:
                port['state_reason_ip'] = self.ip2internal(
                    port['state_reason_ip']
                )
            portid = self.db.execute(
                insert(self.tables.port)
                .values(scan=scanid, **port)
                .returning(self.tables.port.id)
            ).fetchone()[0]
            for script in scripts:
                name, output = script.pop('id'), script.pop('output')
                self.bulk.append(insert(self.tables.script).values(
                    port=portid,
                    name=name,
                    output=output,
                    data=script
                ))
        for trace in host.get('traces', []):
            traceid = self.db.execute(insert(self.tables.trace).values(
                scan=scanid,
                port=trace.get('port'),
                protocol=trace['protocol']
            ).returning(self.tables.trace.id)).fetchone()[0]
            for hop in trace.get('hops'):
                hop['ipaddr'] = self.ip2internal(hop['ipaddr'])
                self.bulk.append(insert(self.tables.hop).values(
                    trace=traceid,
                    ipaddr=self.ip2internal(hop['ipaddr']),
                    ttl=hop["ttl"],
                    rtt=None if hop["rtt"] == '--' else hop["rtt"],
                    host=hop.get("host"),
                    domains=hop.get("domains"),
                ))
        for hostname in host.get('hostnames', []):
            self.bulk.append(insert(self.tables.hostname).values(
                scan=scanid,
                domains=hostname.get('domains'),
                name=hostname.get('name'),
                type=hostname.get('type'),
            ))
        utils.LOGGER.debug("HOST STORED: %r", scanid)
        return scanid

    def store_host(self, host):
        scanid = self._store_host(host)
        insrt = postgresql.insert(self.tables.association_scan_scanfile)
        self.db.execute(insrt
                        .values(scan=scanid,
                                scan_file=utils.decode_hex(host['scanid']))
                        .on_conflict_do_nothing())

    def store_hosts(self, hosts):
        tmp = self.create_tmp_table(self.tables.scan, extracols=[
            Column("scanfileid", ARRAY(LargeBinary(32))),
            Column("categories", ARRAY(String(32))),
            Column("source", String(32)),
            # Column("cpe", postgresql.JSONB),
            # Column("extraports", postgresql.JSONB),
            Column("hostnames", postgresql.JSONB),
            # openports
            # Column("os", postgresql.JSONB),
            Column("ports", postgresql.JSONB),
            # Column("traceroutes", postgresql.JSONB),
        ])
        with ScanCSVFile(hosts, self.ip2internal, tmp) as fdesc:
            self.copy_from(fdesc, tmp.name)


class PostgresDBView(PostgresDBActive, SQLDBView):

    def __init__(self, url):
        super(PostgresDBView, self).__init__(url)

    def _store_host(self, host):
        addr = self.ip2internal(host['addr'])
        info = host.get('infos')
        source = host.get('source', [])
        host_tstart = utils.all2datetime(host['starttime'])
        host_tstop = utils.all2datetime(host['endtime'])
        insrt = postgresql.insert(self.tables.scan)
        scanid, scan_tstop = self.db.execute(
            insrt.values(
                addr=addr,
                source=source,
                info=info,
                time_start=host_tstart,
                time_stop=host_tstop,
                **dict(
                    (key, host.get(key)) for key in
                    ['state', 'state_reason', 'state_reason_ttl']
                    if key in host
                )
            )
            .on_conflict_do_update(
                index_elements=['addr'],
                set_={
                    'source': self.tables.scan.source + insrt.excluded.source,
                    'time_start': func.least(
                        self.tables.scan.time_start,
                        insrt.excluded.time_start,
                    ),
                    'time_stop': func.greatest(
                        self.tables.scan.time_stop,
                        insrt.excluded.time_stop,
                    ),
                },
            )
            .returning(self.tables.scan.id, self.tables.scan.time_stop)
        ).fetchone()
        newest = scan_tstop <= host_tstop
        for category in host.get("categories", []):
            insrt = postgresql.insert(self.tables.category)
            catid = self.db.execute(
                insrt.values(name=category)
                .on_conflict_do_update(
                    index_elements=['name'],
                    set_={'name': insrt.excluded.name}
                )
                .returning(self.tables.category.id)
            ).fetchone()[0]
            self.db.execute(postgresql.insert(
                self.tables.association_scan_category
            ).values(scan=scanid, category=catid)
             .on_conflict_do_nothing())
        for port in host.get('ports', []):
            scripts = port.pop('scripts', [])
            # FIXME: handle screenshots
            for fld in ['screendata', 'screenshot', 'screenwords',
                        'service_method']:
                try:
                    del port[fld]
                except KeyError:
                    pass
            if 'service_servicefp' in port:
                port['service_fp'] = port.pop('service_servicefp')
            if 'state_state' in port:
                port['state'] = port.pop('state_state')
            if 'state_reason_ip' in port:
                port['state_reason_ip'] = self.ip2internal(
                    port['state_reason_ip']
                )
            insrt = postgresql.insert(self.tables.port)
            portid = self.db.execute(
                insrt.values(scan=scanid, **port)
                .on_conflict_do_update(
                    index_elements=['scan', 'port',
                                    'protocol'],
                    set_=dict(
                        scan=scanid,
                        **(port if newest else {})
                    )
                )
                .returning(self.tables.port.id)
            ).fetchone()[0]
            for script in scripts:
                name, output = script.pop('id'), script.pop('output')
                if newest:
                    insrt = postgresql.insert(self.tables.script)
                    self.bulk.append(insrt
                                     .values(
                                         port=portid,
                                         name=name,
                                         output=output,
                                         data=script
                                     )
                                     .on_conflict_do_update(
                                         index_elements=['port', 'name'],
                                         set_={
                                             "output":
                                             insrt.excluded.output,
                                             "data": insrt.excluded.data,
                                         },
                                     ))
                else:
                    insrt = postgresql.insert(self.tables.script)
                    self.bulk.append(insrt
                                     .values(
                                         port=portid,
                                         name=name,
                                         output=output,
                                         data=script
                                     )
                                     .on_conflict_do_nothing())
        for trace in host.get('traces', []):
            traceid = self.db.execute(
                postgresql.insert(self.tables.trace).values(
                    scan=scanid,
                    port=trace.get('port'),
                    protocol=trace['protocol']
                ).on_conflict_do_nothing()
                 .returning(self.tables.trace.id)
            ).fetchone()[0]
            for hop in trace.get('hops'):
                hop['ipaddr'] = self.ip2internal(hop['ipaddr'])
                self.bulk.append(postgresql.insert(self.tables.hop).values(
                    trace=traceid,
                    ipaddr=self.ip2internal(hop['ipaddr']),
                    ttl=hop["ttl"],
                    rtt=None if hop["rtt"] == '--' else hop["rtt"],
                    host=hop.get("host"),
                    domains=hop.get("domains"),
                ))
        for hostname in host.get('hostnames', []):
            self.bulk.append(postgresql.insert(self.tables.hostname).values(
                scan=scanid,
                domains=hostname.get('domains'),
                name=hostname.get('name'),
                type=hostname.get('type'),
            ).on_conflict_do_nothing())
        utils.LOGGER.debug("VIEW STORED: %r", scanid)
        return scanid

    def store_host(self, host):
        self._store_host(host)


class PostgresDBPassive(PostgresDB, SQLDBPassive):

    def __init__(self, url):
        super(PostgresDBPassive, self).__init__(url)
        Index(
            'ix_passive_record', self.tables.passive.addr,
            self.tables.passive.sensor, self.tables.passive.recontype,
            self.tables.passive.port, self.tables.passive.source,
            self.tables.passive.value, self.tables.passive.targetval,
            self.tables.passive.info, unique=True,
            postgresql_where=self.tables.passive.addr != None,
            # noqa: E711 (BinaryExpression)
        )
        Index(
            'ix_passive_record_noaddr', self.tables.passive.sensor,
            self.tables.passive.recontype, self.tables.passive.port,
            self.tables.passive.source, self.tables.passive.value,
            self.tables.passive.targetval, self.tables.passive.info,
            unique=True, postgresql_where=self.tables.passive.addr == None,
            # noqa: E711 (BinaryExpression)
        )

    def _insert_or_update(self, timestamp, values, lastseen=None):
        stmt = postgresql.insert(self.tables.passive).values(values)
        upsert = {
            'firstseen': func.least(
                self.tables.passive.firstseen,
                timestamp,
            ),
            'lastseen': func.greatest(
                self.tables.passive.lastseen,
                lastseen or timestamp,
            ),
            'count': self.tables.passive.count + stmt.excluded.count,
        }
        if values.get('addr'):
            stmt = stmt.on_conflict_do_update(
                index_elements=['addr', 'sensor', 'recontype', 'port',
                                'source', 'value', 'targetval', 'info'],
                index_where=self.tables.passive.addr != None,
                # noqa: E711 (BinaryExpression)
                set_=upsert,
            )
        else:
            stmt = stmt.on_conflict_do_update(
                index_elements=['sensor', 'recontype', 'port',
                                'source', 'value', 'targetval', 'info'],
                index_where=self.tables.passive.addr == None,
                # noqa: E711 (BinaryExpression)
                set_=upsert,
            )
        self.db.execute(stmt)

    def insert_or_update_bulk(self, specs, getinfos=None,
                              separated_timestamps=True):
        """Like `.insert_or_update()`, but `specs` parameter has to be an
        iterable of `(timestamp, spec)` (if `separated_timestamps` is
        True) or `spec` (if it is False) values. This will perform
        PostgreSQL COPY FROM inserts with the major drawback that the
        `getinfos` parameter will be called (if it is not `None`) for
        each spec, even when the spec already exists in the database
        and the call was hence unnecessary.

        It's up to you to decide whether having bulk insert is worth
        it or if you want to go with the regular `.insert_or_update()`
        method.

        """
        more_to_read = True
        tmp = self.create_tmp_table(self.tables.passive)
        if config.DEBUG_DB:
            total_upserted = 0
            total_start_time = time.time()
        while more_to_read:
            if config.DEBUG_DB:
                start_time = time.time()
            with PassiveCSVFile(specs, self.ip2internal, tmp,
                                getinfos=getinfos,
                                separated_timestamps=separated_timestamps,
                                limit=config.POSTGRES_BATCH_SIZE) as fdesc:
                self.copy_from(fdesc, tmp.name)
                more_to_read = fdesc.more_to_read
                if config.DEBUG_DB:
                    count_upserted = fdesc.count
            insrt = postgresql.insert(self.tables.passive)
            self.db.execute(
                insrt.from_select(
                    [column(col) for col in [
                        'addr',
                        # sum / min / max
                        'count', 'firstseen', 'lastseen',
                        # grouped
                        'sensor', 'port', 'recontype', 'source', 'targetval',
                        'value', 'info', 'moreinfo'
                    ]],
                    select([tmp.columns['addr'],
                            func.sum_(tmp.columns['count']),
                            func.min_(tmp.columns['firstseen']),
                            func.max_(tmp.columns['lastseen'])] + [
                                tmp.columns[col] for col in [
                                    'sensor', 'port', 'recontype', 'source',
                                    'targetval', 'value', 'info', 'moreinfo'
                                    ]])\
                    .where(
                        tmp.columns['addr'] != None
                        # noqa: E711 (BinaryExpression)
                    )\
                    .group_by(*(tmp.columns[col] for col in [
                        'addr', 'sensor', 'port', 'recontype', 'source',
                        'targetval', 'value', 'info', 'moreinfo'
                    ]))
                )\
                .on_conflict_do_update(
                    index_elements=['addr', 'sensor', 'recontype', 'port',
                                    'source', 'value', 'targetval', 'info'],
                    index_where=self.tables.passive.addr != None,
                    # noqa: E711 (BinaryExpression)
                    set_={
                        'firstseen': func.least(
                            self.tables.passive.firstseen,
                            insrt.excluded.firstseen,
                        ),
                        'lastseen': func.greatest(
                            self.tables.passive.lastseen,
                            insrt.excluded.lastseen,
                        ),
                        'count':
                        self.tables.passive.count + insrt.excluded.count,
                    },
                )
            )
            self.db.execute(
                insrt.from_select(
                    [column(col) for col in [
                        'addr',
                        # sum / min / max
                        'count', 'firstseen', 'lastseen',
                        # grouped
                        'sensor', 'port', 'recontype', 'source', 'targetval',
                        'value', 'info', 'moreinfo'
                    ]],
                    select([tmp.columns['addr'],
                            func.sum_(tmp.columns['count']),
                            func.min_(tmp.columns['firstseen']),
                            func.max_(tmp.columns['lastseen'])] + [
                                tmp.columns[col] for col in [
                                    'sensor', 'port', 'recontype', 'source',
                                    'targetval', 'value', 'info', 'moreinfo'
                                    ]])\
                    .where(
                        tmp.columns['addr'] == None
                        # noqa: E711 (BinaryExpression)
                    )\
                    .group_by(*(tmp.columns[col] for col in [
                        'addr', 'sensor', 'port', 'recontype', 'source',
                        'targetval', 'value', 'info', 'moreinfo'
                    ]))
                )\
                .on_conflict_do_update(
                    index_elements=['sensor', 'recontype', 'port', 'source',
                                    'value', 'targetval', 'info'],
                    index_where=self.tables.passive.addr == None,
                    # noqa: E711 (BinaryExpression)
                    set_={
                        'firstseen': func.least(
                            self.tables.passive.firstseen,
                            insrt.excluded.firstseen,
                        ),
                        'lastseen': func.greatest(
                            self.tables.passive.lastseen,
                            insrt.excluded.lastseen,
                        ),
                        'count':
                        self.tables.passive.count + insrt.excluded.count,
                    },
                )
            )
            self.db.execute(delete(tmp))
            if config.DEBUG_DB:
                stop_time = time.time()
                time_spent = stop_time - start_time
                total_upserted += count_upserted
                total_time_spent = stop_time - total_start_time
                utils.LOGGER.debug(
                    "DB:PERFORMANCE STATS %s upserts, %f s, %s/s\n"
                    "\ttotal: %s upserts, %f s, %s/s",
                    utils.num2readable(count_upserted), time_spent,
                    utils.num2readable(count_upserted / time_spent),
                    utils.num2readable(total_upserted), total_time_spent,
                    utils.num2readable(total_upserted / total_time_spent),
                )

    def _features_port_get(self, features, flt, yieldall, use_service,
                           use_product, use_version):
        flt = self.flt_and(flt, self.searchport(0, neg=True))
        if use_version:
            fields = [
                cast(self.tables.passive.port, String),
                self.tables.passive.moreinfo.op('->>')('service_name'),
                self.tables.passive.moreinfo.op('->>')('service_product'),
                self.tables.passive.moreinfo.op('->>')('service_version'),
            ]
        elif use_product:
            fields = [
                cast(self.tables.passive.port, String),
                self.tables.passive.moreinfo.op('->>')('service_name'),
                self.tables.passive.moreinfo.op('->>')('service_product'),
            ]
        elif use_service:
            fields = [cast(self.tables.passive.port, String),
                      self.tables.passive.moreinfo.op('->>')('service_name')]
        else:
            fields = [self.tables.passive.port]
        n_features = len(features)
        for addr, cur_features in self.db.execute(
            flt.query(
                select([self.tables.passive.addr,
                        func.array_agg(func.distinct(
                            postgresql.array(fields)
                        ))])
                .group_by(self.tables.passive.addr)
            )
        ):
            currec = [0] * n_features
            for feat in cur_features:
                if use_service:
                    # convert port number back to an integer
                    feat[0] = int(feat[0])
                try:
                    currec[features[tuple(feat)]] = 1
                except KeyError:
                    pass
            yield (addr, currec)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of IVRE.
# Copyright 2011 - 2019 Pierre LALET <pierre.lalet@cea.fr>
#
# IVRE is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IVRE is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with IVRE. If not, see <http://www.gnu.org/licenses/>.

"""
This module is part of IVRE.
Copyright 2011 - 2018 Pierre LALET <pierre.lalet@cea.fr>

This sub-module contains functions used for passive recon.
"""


from datetime import datetime
import hashlib
import re
import struct


from future.utils import viewitems


from ivre import utils, config


SCHEMA_VERSION = 1

# p0f specific

P0F_MODES = {
    'SYN': {
        'options': [],
        'name': 'SYN',
        'filter': 'tcp and tcp[tcpflags] & (tcp-syn|tcp-ack) == 2'
    },
    'SYN+ACK': {
        'options': ['-A'],
        'name': 'SYN+ACK',
        'filter': 'tcp and tcp[tcpflags] & (tcp-syn|tcp-ack) == 18'},
    'RST+': {
        'options': ['-R'],
        'name': 'RST+',
        'filter': 'tcp and tcp[tcpflags] & (tcp-rst) == 4'},
    'ACK': {
        'options': ['-O'],
        'name': 'ACK',
        'filter': 'tcp and tcp[tcpflags] & (tcp-syn|tcp-ack) == 16'}
}

P0F_DIST = re.compile(b'distance ([0-9]+),')


def parse_p0f_line(line, include_port=False, sensor=None, recontype=None):
    line = [line.split(b' - ')[0]] + line.split(b' - ')[1].split(b' -> ')
    if line[1].startswith(b'UNKNOWN '):
        sig = line[1][line[1].index(b'UNKNOWN ') + 8:][1:-1].split(b':')[:6]
        osname, version, dist = b'?', b'?', -1
    else:
        sig = line[1][line[1].index(b' Signature: ') + 12:][
            1:-1].split(b':')[:6]
        if b' (up: ' in line[1]:
            osname = line[1][:line[1].index(b' (up: ')]
        else:
            osname = line[1][:line[1].index(b' Signature: ')]
        osname = osname.split(b' ')
        osname, version = osname[0], b' '.join(osname[1:])
        dist = int(P0F_DIST.search(line[2]).groups()[0])
    # We wildcard any window size which is not Sxxx or Tyyy
    if sig[0][0] not in b'ST':
        sig[0] = b'*'
    spec = {
        'schema_version': SCHEMA_VERSION,
        'addr': line[0][line[0].index(b'> ') + 2:line[0].index(b':')].decode(),
        'distance': dist,
        'value': osname.decode(),
        'version': version.decode(),
        'signature': ':'.join(str(elt) for elt in sig),
    }
    if include_port:
        spec.update({'port': int(line[0][line[0].index(b':') + 1:])})
    if sensor is not None:
        spec['sensor'] = sensor
    if recontype is not None:
        spec['recontype'] = recontype
    return datetime.fromtimestamp(float(line[0][1:line[0].index(b'>')])), spec


# Bro specific

SYMANTEC_UA = re.compile('[a-zA-Z0-9/+]{32,33}AAAAA$')
SYMANTEC_SEP_UA = re.compile(
    '(SEP/[0-9\\.]+),? MID/\\{[A-F0-9]{8}-[A-F0-9]{4}-[A-F0-9]{4}-[A-F0-9]{4}-'
    '[A-F0-9]{12}\\},? SID/[0-9]+(?: SEQ/[0-9]+)?(.*)$'
)
KASPERSKY_UA = re.compile('AAAAA[a-zA-Z0-9_-]{1,2}AB$')
DIGEST_AUTH_INFOS = re.compile('(username|realm|algorithm|qop)=')


def _fix_mysql_banner(match):
    # https://dev.mysql.com/doc/dev/mysql-server/latest/page_protocol_connection_phase_packets_protocol_handshake_v10.html
    plugin_data_len = max(13, struct.unpack('B', match.group(3)[-1:])[0] - 8)
    return (
        match.group(1) + b'\x00\x00' +
        b'\x00' +
        b'\x0a' +
        match.group(2) + b'\x00' +
        b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' +
        match.group(3) +
        match.group(4) +
        b'\x00' * plugin_data_len +
        match.group(5)[plugin_data_len:]
    )


TCP_SERVER_PATTERNS = [
    (re.compile(b'You are user number [0-9]+ of '),
     b'You are user number 1 of '),
    (re.compile(
        b'(?:Mon|Tue|Wed|Thu|Fri|Sat|Sun), [ 0-3]?[0-9] '
        b'(?:Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec) '
        b'[12][0123456789]{3} [0-2][0-9]:[0-9][0-9]:[0-9][0-9]'
    ), b'Thu 1 Jan 1970 00:00:00'),
    (re.compile(b'Local time is now [0-2][0-9]:[0-9][0-9]'),
     b'Local time is now 00:00'),
    (re.compile(
        # MySQL banner
        # https://dev.mysql.com/doc/dev/mysql-server/latest/page_protocol_basic_packets.html
        b'^(.)\x00\x00'  # packet length
        b'\x00'  # packet number
        # https://dev.mysql.com/doc/dev/mysql-server/latest/page_protocol_connection_phase_packets_protocol_handshake_v10.html
        b'\x0a'  # protocol
        b'([3456]\\.[-_~\\.\\+\\w]+)\x00'  # version
        # Thread ID (4) + auth-plugin-data-part-1 (8) + filler (\x00)
        b'.{12}\x00'
        # capabilities 1 (2) + language (1) + status (2) +
        # capabilities 2 (2) + auth_plugin_data_len or 0 (1)
        b'(.{8})'
        b'(.{10})'  # reserved
        b'(.*)',  # auth-plugin-data-part-2 + auth_plugin_name
        re.DOTALL
    ), _fix_mysql_banner),
    # based on Nmap fingerprint
    (re.compile(
        b'^220 ([\\w._-]+) ESMTP '
        b'[a-z]{0,2}[0-9]{1,3}((?:-v6)?)([a-z]{2})[0-9]+[a-z]{3}\\.[0-9]{1,3}'
        b' - gsmtp'
    ), b'220 \\1 ESMTP xx000\\2\\g<3>00000000xxx.000'),
    # OVH
    (re.compile(
        b'220([ -][\\w._-]+) in[0-9]{1,2}($|[\\r\\n])'
    ), b'220\\1 in00\\2'),
    # Outlook
    (re.compile(
        b'^220 ([A-Z]{2})[0-9]([A-Z]{3,4})[0-9]{2}([A-Z]{2,3})[0-9]{3}\\.mail'
        b'\\.protection\\.outlook\\.com '
    ), b'220 \\g<1>0\\g<2>00\\g<3>000.mail.protection.outlook.com '),
    # Yahoo
    (re.compile(
        b'220 mta[0-9]{4}\\.mail\\.([a-z0-9]+)\\.yahoo\\.com ESMTP ready'
    ), b'220 mta0000.mail.\\1.yahoo.com ESMTP ready'),
]


def _split_digest_auth(data):
    """This function handles (Proxy-)Authorization: Digest values"""
    values = []
    curdata = []
    state = 0  # state init
    for char in data:
        if state == 0:
            if char == ',':
                values.append(''.join(curdata).strip())
                curdata = []
            else:
                if char == '"':
                    state = 1  # inside " "
                curdata.append(char)
        elif state == 1:
            if char == '"':
                state = 0
            curdata.append(char)
    values.append(''.join(curdata).strip())
    if state == 1:
        utils.LOGGER.debug("Could not parse Digest auth data [%r]", data)
    return values


def _prepare_rec(spec, ignorenets, neverignore):
    # First of all, let's see if we are supposed to ignore this spec,
    # and if so, do so.
    if 'addr' in spec and \
       spec.get('source') not in neverignore.get(spec['recontype'], []):
        for start, stop in ignorenets.get(spec['recontype'], ()):
            if start <= utils.force_ip2int(spec['addr']) <= stop:
                return None
    # Then, let's clean up the records.
    # Change Symantec's random user agents (matching SYMANTEC_UA) to
    # the constant string 'SymantecRandomUserAgent'.
    if spec['recontype'] == 'HTTP_CLIENT_HEADER' and \
       spec.get('source') == 'USER-AGENT':
        if SYMANTEC_UA.match(spec['value']):
            spec['value'] = 'SymantecRandomUserAgent'
        elif KASPERSKY_UA.match(spec['value']):
            spec['value'] = 'KasperskyWeirdUserAgent'
        else:
            match = SYMANTEC_SEP_UA.match(spec['value'])
            if match is not None:
                spec['value'] = '%s%s' % match.groups()
    # Change any Digest authorization header to remove non-constant
    # information. On one hand we loose the necessary information to
    # try to recover the passwords, but on the other hand we store
    # specs with different challenges but the same username, realm,
    # host and sensor in the same records.
    elif spec['recontype'] in ['HTTP_CLIENT_HEADER',
                               'HTTP_CLIENT_HEADER_SERVER'] and \
        spec.get('source') in ['AUTHORIZATION',
                               'PROXY-AUTHORIZATION']:
        authtype = spec['value'].split(None, 1)[0]
        if authtype.lower() == 'digest':
            try:
                # we only keep relevant info
                value = [val for val in
                         _split_digest_auth(spec['value'][6:].strip())
                         if DIGEST_AUTH_INFOS.match(val)]
                spec['value'] = '%s %s' % (authtype, ','.join(value))
            except Exception:
                utils.LOGGER.warning("Cannot parse digest error for %r", spec,
                                     exc_info=True)
        elif authtype.lower() in ['negotiate', 'kerberos', 'oauth', 'ntlm']:
            spec['value'] = authtype
    # p0f in Bro hack: we use the "source" field to provide the
    # "distance" and "version" values
    elif spec['recontype'] == 'P0F':
        distance, version = spec.pop('source').split('-', 1)
        try:
            spec['distance'] = int(distance)
        except ValueError:
            pass
        if version:
            spec['version'] = version
    # TCP server banners: try to normalize data
    elif spec['recontype'] == 'TCP_SERVER_BANNER':
        newvalue = value = utils.nmap_decode_data(spec['value'])
        for pattern, replace in TCP_SERVER_PATTERNS:
            if pattern.search(newvalue):
                newvalue = pattern.sub(replace, newvalue)
        if newvalue != value:
            spec['value'] = utils.nmap_encode_data(newvalue)
    # SSL_{CLIENT,SERVER} JA3
    elif ((spec['recontype'] == 'SSL_CLIENT' and spec['source'] == 'ja3') or
          (spec['recontype'] == 'SSL_SERVER' and
           spec['source'].startswith('ja3-'))):
        value = spec['value']
        spec.setdefault('infos', {})['raw'] = value
        spec['value'] = hashlib.new("md5", value.encode()).hexdigest()
        if spec['recontype'] == 'SSL_SERVER':
            clientvalue = spec['source'][4:]
            spec['infos'].setdefault('client', {})['raw'] = clientvalue
            spec['source'] = 'ja3-%s' % hashlib.new(
                "md5",
                clientvalue.encode(),
            ).hexdigest()
    # Check DNS Blacklist answer
    elif spec['recontype'] == 'DNS_ANSWER':
        if any(spec['value'].endswith(dnsbl)
               for dnsbl in config.DNS_BLACKLIST_DOMAINS):
            dnsbl_val = spec['value']
            spec['recontype'] = 'DNS_BLACKLIST'
            spec['value'] = spec['addr']
            spec.update({'source': "%s-%s" %
                         (dnsbl_val.split('.', 4)[4], spec['source'])})
            spec['addr'] = '.'.join(dnsbl_val.split('.')[3::-1])
    return spec


def handle_rec(sensor, ignorenets, neverignore,
               # these argmuments are provided by **bro_line
               timestamp=None, uid=None, host=None, srvport=None,
               recon_type=None, source=None, value=None, targetval=None):
    spec = {
        'schema_version': SCHEMA_VERSION,
        'recontype': recon_type,
        'value': value,
    }
    if host is None:
        spec['targetval'] = targetval
    else:
        spec['addr'] = host
    if sensor is not None:
        spec.update({'sensor': sensor})
    if srvport is not None:
        spec.update({'port': srvport})
    if source is not None:
        spec.update({'source': source})
    spec = _prepare_rec(spec, ignorenets, neverignore)
    return timestamp, spec


def _getinfos_http_client_authorization(spec):
    """Extract (for now) the usernames and passwords from Basic
    authorization headers
    """
    infos = {}
    data = spec['value'].split(None, 1)
    if data[1:]:
        if data[0].lower() == b'basic':
            try:
                infos['username'], infos['password'] = utils.decode_b64(
                    data[1].strip()
                ).decode('latin-1').split(':', 1)
            except Exception:
                pass
        elif data[0].lower() == 'digest':
            try:
                infos = dict(
                    value.split('=', 1) if '=' in value else [value, None]
                    for value in _split_digest_auth(data[1].strip())
                )
                for key, value in list(viewitems(infos)):
                    if value.startswith('"') and value.endswith('"'):
                        infos[key] = value[1:-1]
            except Exception:
                pass
    res = {}
    if infos:
        res['infos'] = infos
    return res


def _getinfos_http_server(spec):
    header = utils.nmap_decode_data(spec['value'])
    banner = b"HTTP/1.1 200 OK\r\nServer: " + header + b"\r\n\r\n"
    res = _getinfos_from_banner(banner, probe="GetRequest")
    return res


def _getinfos_dns(spec):
    """Extract domain names in an handy-to-index-and-query form."""
    infos = {}
    fields = {'domain': 'value', 'domaintarget': 'targetval'}
    for field in fields:
        try:
            if fields[field] not in spec:
                continue
            infos[field] = []
            for domain in utils.get_domains(spec[fields[field]]):
                infos[field].append(domain)
            if not infos[field]:
                del infos[field]
        except Exception:
            pass
    res = {}
    if infos:
        res['infos'] = infos
    return res


def _getinfos_dns_blacklist(spec):
    """Extract and properly format DNSBL records."""
    infos = {}
    try:
        if 'source' in spec:
            infos['domain'] = []
            for domain in utils.get_domains(spec['source'].split('-')[-4]):
                infos['domain'].append(domain)
            if not infos['domain']:
                del infos['domain']
    except Exception:
        pass
    res = {}
    if infos:
        res['infos'] = infos
    return res


def _getinfos_sslsrv(spec):
    """Calls a source specific function for SSL_SERVER recontype
records.

    """
    source = spec.get('source')
    if source == 'cert':
        return _getinfos_cert(spec)
    if source.startswith('ja3-'):
        return _getinfos_ja3(spec)
    return {}


def _getinfos_cert(spec):
    """Extract info from a certificate (hash values, issuer, subject,
    algorithm) in an handy-to-index-and-query form.

    """
    # TODO: move to mongodb specific functions.
    try:
        cert = utils.decode_b64(spec['value'].encode())
    except Exception:
        utils.LOGGER.info("Cannot parse certificate for record %r", spec,
                          exc_info=True)
        return {}
    info = utils.get_cert_info(cert)
    res = {}
    if info:
        res['infos'] = info
    return res


def _getinfos_ja3(spec):
    """Extract hashes from JA3 fingerprint strings.

    """
    value = spec['infos']['raw']
    data = value.encode()

    info = dict(((hashtype, hashlib.new(hashtype, data).hexdigest())
                 for hashtype in ['sha1', 'sha256']),
                **spec['infos'])

    if spec.get('recontype') == 'SSL_SERVER':
        clientvalue = spec['infos']['client']['raw']
        clientdata = clientvalue.encode()
        info['client'].update(
            (hashtype, hashlib.new(hashtype, clientdata).hexdigest())
            for hashtype in ['sha1', 'sha256']
        )

    return {'infos': info}


def _getinfos_from_banner(banner, proto="tcp", probe="NULL"):
    infos = utils.match_nmap_svc_fp(banner, proto=proto, probe=probe)
    if not infos:
        return {}
    return {'infos': infos}


def _getinfos_tcp_srv_banner(spec):
    """Extract info from a TCP server banner using Nmap database.

    """
    return _getinfos_from_banner(utils.nmap_decode_data(spec['value']))


def _getinfos_ssh_server(spec):
    """Convert an SSH server banner to a TCP banner and use
_getinfos_tcp_srv_banner()"""
    return _getinfos_from_banner(utils.nmap_decode_data(
        spec['value']
    ) + b'\r\n')


def _getinfos_ssh_hostkey(spec):
    """Parse SSH host keys."""
    infos = {}
    data = utils.nmap_decode_data(spec['value'])
    for hashtype in ['md5', 'sha1', 'sha256']:
        infos[hashtype] = hashlib.new(hashtype, data).hexdigest()
    info = utils.parse_ssh_key(data)
    return {'infos': info}


_GETINFOS_FUNCTIONS = {
    'HTTP_CLIENT_HEADER':
    {'AUTHORIZATION': _getinfos_http_client_authorization,
     'PROXY-AUTHORIZATION': _getinfos_http_client_authorization},
    'HTTP_CLIENT_HEADER_SERVER':
    {'AUTHORIZATION': _getinfos_http_client_authorization,
     'PROXY-AUTHORIZATION': _getinfos_http_client_authorization},
    'HTTP_SERVER_HEADER':
    {'SERVER': _getinfos_http_server},
    'DNS_ANSWER': _getinfos_dns,
    'DNS_BLACKLIST': _getinfos_dns_blacklist,
    'SSL_SERVER': _getinfos_sslsrv,
    'SSL_CLIENT': {'ja3': _getinfos_ja3},
    'TCP_SERVER_BANNER': _getinfos_tcp_srv_banner,
    'SSH_SERVER': _getinfos_ssh_server,
    'SSH_SERVER_HOSTKEY': _getinfos_ssh_hostkey,
}


def getinfos(spec):
    """This functions takes a document from a passive sensor, and
    prepares its 'infos' field (which is not added but returned).

    """
    function = _GETINFOS_FUNCTIONS.get(spec.get('recontype'))
    if isinstance(function, dict):
        function = function.get(spec.get('source'))
    if function is None:
        return {}
    if hasattr(function, '__call__'):
        return function(spec)



#! /usr/bin/env python

# This file is part of IVRE.
# Copyright 2011 - 2018 Pierre LALET <pierre.lalet@cea.fr>
#
# IVRE is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IVRE is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with IVRE. If not, see <http://www.gnu.org/licenses/>.


"""
This program runs a simple httpd server to provide an out-of-the-box
access to the web user interface.

This script should only be used for testing purposes. Production
deployments should use "real" web servers (IVRE has been successfully
tested with both Apache and Nginx).
"""


import os


from bottle import default_app, get, redirect, run, static_file


from ivre.config import DEBUG, WEB_DOKU_PATH, WEB_STATIC_PATH
from ivre.utils import create_argparser
from ivre.web import app as webapp


#
# Index page
#

@get('/')
def server_index():
    """Needed to redirect / to index.html"""
    return redirect('index.html')


#
# Static files
#

@get('/dokuwiki/<filepath:path>')
def server_doku(filepath):
    """This function serves Dokuwiki files as static text files. This is
far from being great...

    """
    filepath = filepath.lower().replace(':', '/')
    if '.' not in os.path.basename(filepath):
        filepath += '.txt'
    return static_file(filepath, root=WEB_DOKU_PATH)


@get('/<filepath:path>')
def server_static(filepath):
    """Serve the static (HTML, JS, CSS, ...) content."""
    return static_file(filepath, root=WEB_STATIC_PATH)


def parse_args():
    """Imports the available module to parse the arguments and return
    the parsed arguments.

    """
    parser, _ = create_argparser(__doc__)
    parser.add_argument('--bind-address', '-b',
                        help='(IP) Address to bind the server to (defaults '
                        'to 127.0.0.1).',
                        default="127.0.0.1")
    parser.add_argument('--port', '-p', type=int, default=80,
                        help='(TCP) Port to use (defaults to 80)')
    return parser.parse_args()


def main():
    """Function run when the tool is called."""
    print(__doc__)
    args = parse_args()
    application = default_app()
    application.mount('/cgi/', webapp.application)
    run(host=args.bind_address, port=args.port, debug=DEBUG)



#! /usr/bin/env python

# This file is part of IVRE.
# Copyright 2011 - 2018 Pierre LALET <pierre.lalet@cea.fr>
#
# IVRE is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IVRE is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with IVRE. If not, see <http://www.gnu.org/licenses/>.

"""Put selected results in views."""

from __future__ import print_function
try:
    import argparse
except ImportError:
    from itertools import chain
    import optparse
    USING_ARGPARSE = False
else:
    USING_ARGPARSE = True
import os
import sys

from builtins import input

from ivre import graphroute
from ivre.db import db
from ivre.nmapout import displayhosts
from ivre.activecli import display_short, display_distinct, \
    displayfunction_json, displayfunction_honeyd, displayfunction_nmapxml, \
    displayfunction_gnmap, displayfunction_graphroute, \
    displayfunction_explain, displayfunction_remove, displayfunction_csv
from ivre.utils import display_top, CLI_ARGPARSER


def main():
    if USING_ARGPARSE:
        parser = argparse.ArgumentParser(
            description='Print out views.',
            parents=[db.view.argparser, CLI_ARGPARSER])
    else:
        parser = optparse.OptionParser(
            description='Print out views.')
        for args, kargs in chain(db.view.argparser.args, CLI_ARGPARSER):
            parser.add_option(*args, **kargs)
        parser.parse_args_orig = parser.parse_args

        def my_parse_args():
            res = parser.parse_args_orig()
            res[0].ensure_value('ips', res[1])
            return res[0]
        parser.parse_args = my_parse_args
        parser.add_argument = parser.add_option

    flt = db.view.flt_empty

    parser.add_argument('--verbose', '-v', action='store_true',
                        help='Print out formatted results.')
    parser.add_argument('--no-screenshots', action='store_true',
                        help='When used with --json, do not output '
                        'screenshots data.')
    parser.add_argument('--honeyd', action='store_true',
                        help='Output results as a honeyd config file.')
    parser.add_argument('--nmap-xml', action='store_true',
                        help='Output results as a nmap XML output file.')
    parser.add_argument('--gnmap', action='store_true',
                        help='Output results as a nmap grepable output file.')
    parser.add_argument(
        '--graphroute',
        choices=["dot", "rtgraph3d"] if graphroute.HAVE_DBUS else ["dot"],
        help='Create a graph from traceroute results. '
        'dot: output result as Graphviz "dot" format to stdout.'
        '%s' % (" rtgraph3d: send results to rtgraph3d."
                if graphroute.HAVE_DBUS else "")
    )
    parser.add_argument('--graphroute-cluster', choices=['AS', 'Country'],
                        help='Cluster IP according to the specified criteria'
                        '(only for --graphroute dot)')
    if graphroute.HAVE_DBUS:
        parser.add_argument('--graphroute-dont-reset', action='store_true',
                            help='Do NOT reset graph (only for '
                            '--graphroute rtgraph3d)')
    parser.add_argument('--graphroute-include', choices=['last-hop', 'target'],
                        help='How far should graphroute go? Default if to '
                        'exclude the last hop and the target for each result.')
    parser.add_argument('--top', metavar='FIELD / ~FIELD',
                        help='Output most common (least common: ~) values for '
                        'FIELD, by default 10, use --limit to change that, '
                        '--limit 0 means unlimited.')
    parser.add_argument('--csv', metavar='TYPE',
                        help='Output result as a CSV file',
                        choices=['ports', 'hops'])
    parser.add_argument('--csv-separator', metavar='SEPARATOR',
                        default=",",
                        help='Select separator for --csv output')
    parser.add_argument('--csv-add-infos', action='store_true',
                        help="Include country_code and as_number"
                        "fields to CSV file")
    parser.add_argument('--csv-na-str', default="NA",
                        help='String to use for "Not Applicable" value '
                        '(defaults to "NA")')

    args = parser.parse_args()

    flt = db.view.parse_args(args)

    if args.init:
        if os.isatty(sys.stdin.fileno()):
            sys.stdout.write(
                'This will remove any view in your database. Process ? [y/N] '
            )
            ans = input()
            if ans.lower() not in ['y', 'yes']:
                exit(0)
        db.view.init()
        exit(0)

    if args.top is not None:
        display_top(db.view, args.top, flt, args.limit)
        sys.exit(0)
    if args.sort is not None:
        sortkeys = [(field[1:], -1) if field.startswith('~') else (field, 1)
                    for field in args.sort]
    else:
        sortkeys = []
    if args.short:
        display_short(db.view, flt, sortkeys, args.limit, args.skip)
        sys.exit(0)
    if args.distinct is not None:
        display_distinct(db.view, args.distinct, flt, sortkeys,
                         args.limit, args.skip)
        sys.exit(0)
    if args.explain:
        displayfunction_explain(flt, db.view)
        sys.exit(0)
    if args.json:
        def displayfunction(x):
            return displayfunction_json(
                x, db.view, args.no_screenshots
            )
    elif args.honeyd:
        displayfunction = displayfunction_honeyd
    elif args.nmap_xml:
        displayfunction = displayfunction_nmapxml
    elif args.gnmap:
        displayfunction = displayfunction_gnmap
    elif args.graphroute is not None:
        def displayfunction(x):
            return displayfunction_graphroute(
                x, args.graphroute, args.graphroute_include,
                args.graphroute_dont_reset
            )
    elif args.delete:
        def displayfunction(x):
            return displayfunction_remove(x, db.view)
    elif args.csv is not None:
        def displayfunction(x):
            return displayfunction_csv(
                x, args.csv, args.csv_separator, args.csv_na_str,
                args.csv_add_infos
            )
    else:

        def displayfunction(cursor):
            displayhosts(cursor, out=sys.stdout)

    if args.update_schema:
        db.db.nmap.migrate_schema(args.version)
    elif args.count:
        sys.stdout.write(
            str(db.view.count(flt)) + '\n'
        )
    else:
        kargs = {}
        if args.limit is not None:
            kargs["limit"] = args.limit
        if args.skip is not None:
            kargs["skip"] = args.skip
        if sortkeys:
            kargs["sort"] = sortkeys
        cursor = db.view.get(flt, **kargs)
        displayfunction(cursor)
        sys.exit(0)



#! /usr/bin/env python

# This file is part of IVRE.
# Copyright 2011 - 2018 Pierre LALET <pierre.lalet@cea.fr>
#
# IVRE is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IVRE is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with IVRE. If not, see <http://www.gnu.org/licenses/>.


"""
This module provides the dynamic (server-side) part of the Web
interface.

It is used by the integrated web server (ivre httpd) and by the WSGI
application.
"""


from collections import namedtuple
from functools import wraps
import json
import os
import tempfile


from bottle import abort, request, response, Bottle
from future.utils import viewitems


from ivre import config, utils, VERSION
from ivre.db import db
from ivre.web import utils as webutils


application = Bottle()


#
# Utils
#

def check_referer(func):
    """"Wrapper for route functions to implement a basic anti-CSRF check
based on the Referer: header.

    It will abort if the referer is invalid.

    """

    if config.WEB_ALLOWED_REFERERS is False:
        return func

    def _die(referer):
        utils.LOGGER.critical("Invalid Referer header [%r]", referer)
        response.set_header('Content-Type', 'application/javascript')
        response.status = '400 Bad Request'
        return webutils.js_alert(
            "referer", "error",
            "Invalid Referer header. Check your configuration."
        )

    @wraps(func)
    def _newfunc(*args, **kargs):
        referer = request.headers.get('Referer')
        if not referer:
            return _die(referer)
        if config.WEB_ALLOWED_REFERERS is None:
            base_url = '/'.join(request.url.split('/', 3)[:3]) + '/'
            if referer.startswith(base_url):
                return func(*args, **kargs)
        elif referer in config.WEB_ALLOWED_REFERERS:
            return func(*args, **kargs)
        return _die(referer)

    return _newfunc


#
# Configuration
#

@application.get('/config')
@check_referer
def get_config():
    """This function returns JS code to set client-side config values."""
    response.set_header('Content-Type', 'application/javascript')
    for key, value in [
            ("notesbase", config.WEB_NOTES_BASE),
            ("dflt_limit", config.WEB_LIMIT),
            ("warn_dots_count", config.WEB_WARN_DOTS_COUNT),
            ("publicsrv", config.WEB_PUBLIC_SRV),
            ("uploadok", config.WEB_UPLOAD_OK),
            ("flow_time_precision", config.FLOW_TIME_PRECISION),
            ("version", VERSION),
    ]:
        yield "config.%s = %s;\n" % (key, json.dumps(value))


#
# /nmap/
#

FilterParams = namedtuple("flt_params", ['flt', 'sortby', 'unused',
                                         'skip', 'limit', 'callback',
                                         'ipsasnumbers', 'datesasstrings'])


def get_nmap_base(dbase):
    response.set_header('Content-Type', 'application/javascript')
    query = webutils.query_from_params(request.params)
    flt, sortby, unused, skip, limit = webutils.flt_from_query(dbase, query)
    if limit is None:
        limit = config.WEB_LIMIT
    if config.WEB_MAXRESULTS is not None:
        limit = min(limit, config.WEB_MAXRESULTS)
    callback = request.params.get("callback")
    # type of result
    ipsasnumbers = request.params.get("ipsasnumbers")
    datesasstrings = request.params.get("datesasstrings")
    if callback is None:
        response.set_header('Content-Disposition',
                            'attachment; filename="IVRE-results.json"')
    return FilterParams(flt, sortby, unused, skip, limit, callback,
                        ipsasnumbers, datesasstrings)


@application.get(
    '/<subdb:re:scans|view>/<action:re:'
    'onlyips|ipsports|timeline|coordinates|countopenports|diffcats>'
)
@check_referer
def get_nmap_action(subdb, action):
    subdb = db.view if subdb == 'view' else db.nmap
    flt_params = get_nmap_base(subdb)
    preamble = "[\n"
    postamble = "]\n"
    if action == "timeline":
        result, count = subdb.get_open_port_count(flt_params.flt)
        if request.params.get("modulo") is None:
            def r2time(r):
                return int(utils.datetime2timestamp(r['starttime']))
        else:
            def r2time(r):
                return (int(utils.datetime2timestamp(r['starttime']))
                        % int(request.params.get("modulo")))
        if flt_params.ipsasnumbers:
            def r2res(r):
                return [r2time(r), utils.ip2int(r['addr']),
                        r['openports']['count']]
        else:
            def r2res(r):
                return [r2time(r), r['addr'], r['openports']['count']]
    elif action == "coordinates":
        def r2res(r):
            return {
                "type": "Point",
                "coordinates": r['_id'][::-1],
                "properties": {"count": r['count']},
            }
        preamble = '{"type": "GeometryCollection", "geometries": [\n'
        postamble = ']}\n'
        result = list(subdb.getlocations(flt_params.flt))
        count = len(result)
    elif action == "countopenports":
        result, count = subdb.get_open_port_count(flt_params.flt)
        if flt_params.ipsasnumbers:
            def r2res(r):
                return [utils.ip2int(r['addr']), r['openports']['count']]
        else:
            def r2res(r):
                return [r['addr'], r['openports']['count']]
    elif action == "ipsports":
        result, count = subdb.get_ips_ports(flt_params.flt)
        if flt_params.ipsasnumbers:
            def r2res(r):
                return [
                    utils.ip2int(r['addr']),
                    [[p['port'], p['state_state']]
                     for p in r.get('ports', [])
                     if 'state_state' in p]
                ]
        else:
            def r2res(r):
                return [
                    r['addr'],
                    [[p['port'], p['state_state']]
                     for p in r.get('ports', [])
                     if 'state_state' in p]
                ]
    elif action == "onlyips":
        result, count = subdb.get_ips(flt_params.flt)
        if flt_params.ipsasnumbers:
            def r2res(r):
                return utils.ip2int(r['addr'])
        else:
            def r2res(r):
                return r['addr']
    elif action == "diffcats":
        def r2res(r):
            return r
        if request.params.get("onlydiff"):
            output = subdb.diff_categories(request.params.get("cat1"),
                                           request.params.get("cat2"),
                                           flt=flt_params.flt,
                                           include_both_open=False)
        else:
            output = subdb.diff_categories(request.params.get("cat1"),
                                           request.params.get("cat2"),
                                           flt=flt_params.flt)
        count = 0
        result = {}
        if flt_params.ipsasnumbers:
            for res in output:
                result.setdefault(res["addr"], []).append([res['port'],
                                                           res['value']])
                count += 1
        else:
            for res in output:
                result.setdefault(utils.int2ip(res["addr"]),
                                  []).append([res['port'], res['value']])
                count += 1
        result = viewitems(result)
    if flt_params.callback is not None:
        if count >= config.WEB_WARN_DOTS_COUNT:
            yield (
                'if(confirm("You are about to ask your browser to display %d '
                'dots, which is a lot and might slow down, freeze or crash '
                'your browser. Do you want to continue?")) {\n' % count
            )
        yield '%s(\n' % flt_params.callback
    yield preamble

    # hack to avoid a trailing comma
    result = iter(result)
    try:
        rec = next(result)
    except StopIteration:
        pass
    else:
        yield json.dumps(r2res(rec))
        for rec in result:
            yield ",\n" + json.dumps(r2res(rec))
        yield "\n"

    yield postamble
    if flt_params.callback is not None:
        yield ");"
        if count >= config.WEB_WARN_DOTS_COUNT:
            yield '}\n'
    else:
        yield "\n"


@application.get('/<subdb:re:scans|view>/count')
@check_referer
def get_nmap_count(subdb):
    subdb = db.view if subdb == 'view' else db.nmap
    flt_params = get_nmap_base(subdb)
    count = subdb.count(flt_params.flt)
    if flt_params.callback is None:
        return "%d\n" % count
    return "%s(%d);\n" % (flt_params.callback, count)


@application.get('/<subdb:re:scans|view>/top/<field:path>')
@check_referer
def get_nmap_top(subdb, field):
    subdb = db.view if subdb == 'view' else db.nmap
    flt_params = get_nmap_base(subdb)
    if field[0] in '-!':
        field = field[1:]
        least = True
    else:
        least = False
    topnbr = 15
    if ':' in field:
        field, topnbr = field.rsplit(':', 1)
        try:
            topnbr = int(topnbr)
        except ValueError:
            field = '%s:%s' % (field, topnbr)
            topnbr = 15
    if flt_params.callback is None:
        yield "[\n"
    else:
        yield "%s([\n" % flt_params.callback
    # hack to avoid a trailing comma
    cursor = iter(subdb.topvalues(
        field, flt=flt_params.flt, least=least, topnbr=topnbr,
    ))
    try:
        rec = next(cursor)
    except StopIteration:
        pass
    else:
        yield json.dumps({"label": rec['_id'], "value": rec['count']})
        for rec in cursor:
            yield ",\n%s" % json.dumps({"label": rec['_id'],
                                        "value": rec['count']})
    if flt_params.callback is None:
        yield "\n]\n"
    else:
        yield "\n]);\n"


@application.get('/<subdb:re:scans|view>')
@check_referer
def get_nmap(subdb):
    subdb = db.view if subdb == 'view' else db.nmap
    flt_params = get_nmap_base(subdb)
    # PostgreSQL: the query plan if affected by the limit and gives
    # really poor results. This is a temporary workaround (look for
    # XXX-WORKAROUND-PGSQL).
    # result = subdb.get(flt_params.flt, limit=flt_params.limit,
    #                    skip=flt_params.skip, sort=flt_params.sortby)
    result = subdb.get(flt_params.flt, skip=flt_params.skip,
                       sort=flt_params.sortby)

    if flt_params.unused:
        msg = 'Option%s not understood: %s' % (
            's' if len(flt_params.unused) > 1 else '',
            ', '.join(flt_params.unused),
        )
        if flt_params.callback is not None:
            yield webutils.js_alert("param-unused", "warning", msg)
        utils.LOGGER.warning(msg)
    elif flt_params.callback is not None:
        yield webutils.js_del_alert("param-unused")

    if config.DEBUG:
        msg1 = "filter: %s" % subdb.flt2str(flt_params.flt)
        msg2 = "user: %r" % webutils.get_user()
        utils.LOGGER.debug(msg1)
        utils.LOGGER.debug(msg2)
        if flt_params.callback is not None:
            yield webutils.js_alert("filter", "info", msg1)
            yield webutils.js_alert("user", "info", msg2)

    version_mismatch = {}
    if flt_params.callback is None:
        yield "[\n"
    else:
        yield "%s([\n" % flt_params.callback
    # XXX-WORKAROUND-PGSQL
    # for rec in result:
    for i, rec in enumerate(result):
        for fld in ['_id', 'scanid']:
            try:
                del rec[fld]
            except KeyError:
                pass
        if not flt_params.ipsasnumbers:
            rec['addr'] = utils.force_int2ip(rec['addr'])
        for field in ['starttime', 'endtime']:
            if field in rec:
                if not flt_params.datesasstrings:
                    rec[field] = int(utils.datetime2timestamp(rec[field]))
        for port in rec.get('ports', []):
            if 'screendata' in port:
                port['screendata'] = utils.encode_b64(port['screendata'])
            for script in port.get('scripts', []):
                if "masscan" in script:
                    try:
                        del script['masscan']['raw']
                    except KeyError:
                        pass
        if not flt_params.ipsasnumbers:
            if 'traces' in rec:
                for trace in rec['traces']:
                    trace['hops'].sort(key=lambda x: x['ttl'])
                    for hop in trace['hops']:
                        hop['ipaddr'] = utils.force_int2ip(hop['ipaddr'])
        yield "%s\t%s" % ('' if i == 0 else ',\n',
                          json.dumps(rec, default=utils.serialize))
        check = subdb.cmp_schema_version_host(rec)
        if check:
            version_mismatch[check] = version_mismatch.get(check, 0) + 1
        # XXX-WORKAROUND-PGSQL
        if i + 1 >= flt_params.limit:
            break
    if flt_params.callback is None:
        yield "\n]\n"
    else:
        yield "\n]);\n"

    messages = {
        1: lambda count: ("%d document%s displayed %s out-of-date. Please run "
                          "the following command: 'ivre scancli "
                          "--update-schema;" % (count,
                                                's' if count > 1 else '',
                                                'are' if count > 1 else 'is')),
        -1: lambda count: ('%d document%s displayed ha%s been inserted by '
                           'a more recent version of IVRE. Please update '
                           'IVRE!' % (count, 's' if count > 1 else '',
                                      've' if count > 1 else 's')),
    }
    for mismatch, count in viewitems(version_mismatch):
        message = messages[mismatch](count)
        if flt_params.callback is not None:
            yield webutils.js_alert(
                "version-mismatch-%d" % ((mismatch + 1) // 2),
                "warning", message
            )
        utils.LOGGER.warning(message)


#
# Upload scans
#

def parse_form():
    categories = request.forms.get("categories")
    categories = (set(categories.split(',')) if categories else set())
    source = request.forms.get("source")
    if not source:
        utils.LOGGER.critical("source is mandatory")
        abort(400, "ERROR: source is mandatory\n")
    files = request.files.getall("result")
    if config.WEB_PUBLIC_SRV:
        if webutils.get_user() is None:
            utils.LOGGER.critical("username is mandatory on public instances")
            abort(400, "ERROR: username is mandatory on public instances")
        if request.forms.get('public') == 'on':
            categories.add('Shared')
        user = webutils.get_anonymized_user()
        categories.add(user)
        source = "%s-%s" % (user, source)
    return (request.forms.get('referer'), source, categories, files)


def import_files(subdb, source, categories, files):
    count = 0
    categories = list(categories)
    for fileelt in files:
        fdesc = tempfile.NamedTemporaryFile(delete=False)
        fileelt.save(fdesc)
        try:
            if subdb.store_scan(fdesc.name, categories=categories,
                                source=source):
                count += 1
                os.unlink(fdesc.name)
            else:
                utils.LOGGER.warning("Could not import %s", fdesc.name)
        except Exception:
            utils.LOGGER.warning("Could not import %s", fdesc.name,
                                 exc_info=True)
    return count


@application.post('/<subdb:re:scans|view>')
@check_referer
def post_nmap(subdb):
    referer, source, categories, files = parse_form()
    count = import_files(db.view if subdb == 'view' else db.nmap, source,
                         categories, files)
    if request.params.get("output") == "html":
        response.set_header('Refresh', '5;url=%s' % referer)
        return """<html>
  <head>
    <title>IVRE Web UI</title>
  </head>
  <body style="padding-top: 2%%; padding-left: 2%%">
    <h1>%d result%s uploaded</h1>
  </body>
</html>""" % (count, 's' if count > 1 else '')
    return {"count": count}


#
# /flow/
#

@application.get('/flows')
@check_referer
def get_flow():
    callback = request.params.get("callback")
    action = request.params.get("action", "")
    if callback is None:
        response.set_header('Content-Disposition',
                            'attachment; filename="IVRE-results.json"')
    else:
        yield callback + "(\n"
    utils.LOGGER.debug("Params: %r", dict(request.params))
    query = json.loads(request.params.get('q', '{}'))
    limit = query.get("limit", config.WEB_GRAPH_LIMIT)
    skip = query.get("skip", config.WEB_GRAPH_LIMIT)
    mode = query.get("mode", "default")
    count = query.get("count", False)
    orderby = query.get("orderby", None)
    timeline = query.get("timeline", False)
    utils.LOGGER.debug("Action: %r, Query: %r", action, query)
    if action == "details":
        # TODO: error
        if "Host" in query["labels"]:
            res = db.flow.host_details(query["id"])
        else:
            res = db.flow.flow_details(query["id"])
    else:
        cquery = db.flow.from_filters(query, limit=limit, skip=skip,
                                      orderby=orderby, mode=mode,
                                      timeline=timeline)
        if count:
            res = db.flow.count(cquery)
        else:
            res = db.flow.to_graph(cquery)
    yield json.dumps(res, default=utils.serialize)
    if callback is not None:
        yield ");\n"


#
# /ipdata/
#

@application.get('/ipdata/<addr>')
@check_referer
def get_ipdata(addr):
    callback = request.params.get("callback")
    result = json.dumps(db.data.infos_byip(addr))
    if callback is None:
        return result + '\n'
    return "%s(%s);\n" % (callback, result)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of IVRE.
# Copyright 2011 - 2018 Pierre LALET <pierre.lalet@cea.fr>
#
# IVRE is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IVRE is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with IVRE. If not, see <http://www.gnu.org/licenses/>.

"""
This program runs scans and produces output files importable with
ivre scan2db.
"""


from __future__ import print_function
import atexit
import fcntl
from functools import reduce
import multiprocessing
import os
import re
import resource
import select
import shutil
import subprocess
import sys
import termios
import time


from future.utils import viewitems


import ivre.agent
import ivre.geoiputils
import ivre.utils
import ivre.target
import ivre.nmapopt


if sys.version_info >= (2, 7):
    import functools
    USE_PARTIAL = True
else:
    # Python version <= 2.6:
    # see http://bugs.python.org/issue5228
    # multiprocessing not compatible with functools.partial
    USE_PARTIAL = False
    # Also Python version <= 2.6: cannot use a function defined in
    # another function in a multiprocessing.Pool.imap()

    def _call_nmap_single_tuple(args):
        return _call_nmap_single(*args)


STATUS_NEW = 0
STATUS_DONE_UP = 1
STATUS_DONE_DOWN = 2
STATUS_DONE_UNKNOWN = 3

NMAP_LIMITS = {}


def setnmaplimits():
    """Enforces limits from NMAP_LIMITS global variable."""
    for limit, value in viewitems(NMAP_LIMITS):
        resource.setrlimit(limit, value)


class XmlProcess(object):
    addrrec = re.compile(b'<address\\s+addr="([0-9\\.]+)" addrtype="ipv4"/>')

    def target_status(self, _):
        return STATUS_NEW


class XmlProcessTest(XmlProcess):

    def process(self, fdesc):
        data = fdesc.read()
        if not data:
            return False
        for addr in self.addrrec.finditer(data):
            print("Read address", addr.groups()[0].decode())
        return True


class XmlProcessWritefile(XmlProcess):
    statusline = re.compile(b'<task(begin|end|progress).*/>\n')
    status_up = b'<status state="up"'
    status_down = b'<status state="down"'
    hostbegin = re.compile(b'<host[\\s>]')
    status_paths = {
        'up': STATUS_DONE_UP,
        'down': STATUS_DONE_DOWN,
        'unknown': STATUS_DONE_UNKNOWN,
    }

    def __init__(self, path, fulloutput=False):
        self.path = path
        self.starttime = int(time.time() * 1000000)
        self.data = b''
        self.isstarting = True
        self.startinfo = b''
        ivre.utils.makedirs(self.path)
        self.scaninfo = open('%sscaninfo.%d' % (self.path,
                                                self.starttime),
                             'wb')
        if fulloutput:
            self.has_fulloutput = True
            self.fulloutput = open('%sfulloutput.%d' % (self.path,
                                                        self.starttime),
                                   'wb')
        else:
            self.has_fulloutput = False

    def process(self, fdesc):
        newdata = fdesc.read()
        # print("READ", len(newdata), "bytes")
        if not newdata:
            self.scaninfo.write(self.data)
            self.scaninfo.close()
            if self.has_fulloutput:
                self.fulloutput.close()
            return False
        if self.has_fulloutput:
            self.fulloutput.write(newdata)
            self.fulloutput.flush()
        self.data += newdata
        while b'</host>' in self.data:
            hostbeginindex = self.data.index(
                self.hostbegin.search(self.data).group())
            self.scaninfo.write(self.data[:hostbeginindex])
            self.scaninfo.flush()
            if self.isstarting:
                self.startinfo += self.statusline.sub(
                    b'', self.data[:hostbeginindex],
                )
                self.isstarting = False
            self.data = self.data[hostbeginindex:]
            hostrec = self.data[:self.data.index(b'</host>') + 7]
            try:
                addr = self.addrrec.search(hostrec).groups()[0]
            except Exception:
                ivre.utils.LOGGER.warning("Exception for record %r", hostrec,
                                          exc_info=True)
            if self.status_up in hostrec:
                status = 'up'
            elif self.status_down in hostrec:
                status = 'down'
            else:
                status = 'unknown'
            outfile = self.path + status + \
                '/' + addr.decode().replace('.', '/') + '.xml'
            ivre.utils.makedirs(os.path.dirname(outfile))
            with open(outfile, 'wb') as out:
                # out.write(b'<scaninfo starttime="%d" />\n' % starttime)
                out.write(self.startinfo)
                out.write(hostrec)
                out.write(b'\n</nmaprun>\n')
            self.data = self.data[self.data.index(b'</host>') + 7:]
            if self.data.startswith(b'\n'):
                self.data = self.data[1:]
        return True

    def target_status(self, target):
        for status, statuscode in viewitems(self.status_paths):
            try:
                os.stat(os.path.join(self.path, status,
                                     target.replace('.', '/') + '.xml'))
                return statuscode
            except OSError:
                pass
        return STATUS_NEW


def restore_echo():
    """Hack for https://stackoverflow.com/questions/6488275 equivalent
    issue with Nmap (from
    http://stackoverflow.com/a/8758047/3223422)

    """
    try:
        fdesc = sys.stdin.fileno()
    except ValueError:
        return
    try:
        attrs = termios.tcgetattr(fdesc)
    except termios.error:
        return
    attrs[3] = attrs[3] | termios.ECHO
    termios.tcsetattr(fdesc, termios.TCSADRAIN, attrs)


def call_nmap(options, xmlprocess, targets,
              accept_target_status=None):
    if accept_target_status is None:
        accept_target_status = [STATUS_NEW]
    options += ['-oX', '-', '-iL', '-']
    proc = subprocess.Popen(options, preexec_fn=setnmaplimits,
                            stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    procout = proc.stdout.fileno()
    procoutfl = fcntl.fcntl(procout, fcntl.F_GETFL)
    fcntl.fcntl(procout, fcntl.F_SETFL, procoutfl | os.O_NONBLOCK)
    toread = [proc.stdout]
    towrite = [proc.stdin]
    targiter = targets.__iter__()
    while toread:
        # print("ENTERING SELECT")
        rlist, wlist = select.select(toread, towrite, [])[:2]
        # print("LEAVING SELECT", rlist, wlist)
        for rfdesc in rlist:
            # print("PROCESSING DATA")
            if not xmlprocess.process(rfdesc):
                print("NO MORE DATA TO PROCESS")
                rfdesc.close()
                toread.remove(rfdesc)
        for wfdesc in wlist:
            try:
                naddr = ivre.utils.int2ip(next(targiter))
                while xmlprocess.target_status(
                        naddr) not in accept_target_status:
                    naddr = ivre.utils.int2ip(next(targiter))
                print("ADDING TARGET", end=' ')
                print(targiter.nextcount, end=' ')
                if hasattr(targets, "targetcount"):
                    print('/', targets.targetscount, end=' ')
                print(":", naddr)
                wfdesc.write(naddr.encode() + b'\n')
                wfdesc.flush()
            except StopIteration:
                print("WROTE ALL TARGETS")
                wfdesc.close()
                towrite.remove(wfdesc)
            except IOError:
                print("ERROR: NMAP PROCESS IS DEAD")
                return -1
    proc.wait()
    return 0


def _call_nmap_single(maincategory, options,
                      accept_target_status, target):
    target = ivre.utils.int2ip(target)
    outfile = 'scans/%s/%%s/%s.xml' % (maincategory, target.replace('.', '/'))
    if STATUS_DONE_UP not in accept_target_status:
        try:
            os.stat(outfile % 'up')
            return
        except OSError:
            pass
    if STATUS_DONE_DOWN not in accept_target_status:
        try:
            os.stat(outfile % 'down')
            return
        except OSError:
            pass
    if STATUS_DONE_UNKNOWN not in accept_target_status:
        try:
            os.stat(outfile % 'unknown')
            return
        except OSError:
            pass
    ivre.utils.makedirs(os.path.dirname(outfile % 'current'))
    subprocess.call(options + ['-oX', outfile % 'current', target],
                    preexec_fn=setnmaplimits)
    resdata = open(outfile % 'current', 'rb').read()
    if b'<status state="up"' in resdata:
        outdir = 'up'
    elif b'<status state="down"' in resdata:
        outdir = 'down'
    else:
        outdir = 'unknown'
    ivre.utils.makedirs(os.path.dirname(outfile % outdir))
    shutil.move(outfile % 'current', outfile % outdir)


def main():
    atexit.register(restore_echo)
    accept_target_status = set([STATUS_NEW])
    try:
        import argparse
        parser = argparse.ArgumentParser(
            description='Run massive nmap scans.',
            parents=[ivre.target.ARGPARSER,
                     ivre.nmapopt.ARGPARSER])
        using_argparse = True
    except ImportError:
        import optparse
        parser = optparse.OptionParser(
            description='Run massive nmap scans.')
        for parent in [ivre.target.ARGPARSER, ivre.nmapopt.ARGPARSER]:
            for args, kargs in parent.args:
                parser.add_option(*args, **kargs)
        parser.parse_args_orig = parser.parse_args
        parser.parse_args = lambda: parser.parse_args_orig()[0]
        parser.add_argument = parser.add_option
        using_argparse = False
    parser.add_argument('--output',
                        choices=['XML', 'XMLFull', 'XMLFork', 'Test',
                                 'Count', 'List', 'ListAll',
                                 'ListAllRand', 'ListCIDRs',
                                 'CommandLine', 'Agent'],
                        default='XML',
                        help='select output method for scan results')
    parser.add_argument('--processes', metavar='COUNT', type=int, default=30,
                        help='run COUNT nmap processes in parallel '
                        '(when --output=XMLFork)')
    parser.add_argument('--nmap-max-cpu', metavar='TIME', type=int,
                        help="maximum amount of CPU time (in seconds) "
                        "per nmap process")
    parser.add_argument('--nmap-max-heap-size', metavar='SIZE', type=int,
                        help="maximum size (in bytes) of each nmap "
                        "process's heap")
    parser.add_argument('--nmap-max-stack-size', metavar='SIZE', type=int,
                        help="maximum size (in bytes) of each nmap "
                        "process's stack")
    if using_argparse:
        parser.add_argument('--again', nargs='+',
                            choices=['up', 'down', 'unknown', 'all'],
                            help='select status of targets to scan again')
    else:
        parser.add_argument('--again',
                            choices=['up', 'down', 'unknown', 'all'],
                            help='select status of targets to scan again')
    args = parser.parse_args()
    if args.output == 'CommandLine':
        print("Command line to run a scan with template "
              "%s" % args.nmap_template)
        print("    %s" % ivre.nmapopt.build_nmap_commandline(
            template=args.nmap_template,
        ))
        exit(0)
    if args.output == 'Agent':
        sys.stdout.write(ivre.agent.build_agent(template=args.nmap_template))
        exit(0)
    if args.output == 'Count':
        if args.country is not None:
            print('%s has %d IPs.' % (
                args.country,
                ivre.geoiputils.count_ips_by_country(args.country)
            ))
            exit(0)
        if args.region is not None:
            print('%s / %s has %d IPs.' % (
                args.region[0], args.region[1],
                ivre.geoiputils.count_ips_by_region(*args.region),
            ))
            exit(0)
        if args.city is not None:
            print('%s / %s has %d IPs.' % (
                args.city[0], args.city[1],
                ivre.geoiputils.count_ips_by_city(*args.city),
            ))
            exit(0)
        if args.asnum is not None:
            print('AS%d has %d IPs.' % (
                args.asnum,
                ivre.geoiputils.count_ips_by_asnum(args.asnum)
            ))
            exit(0)
        if args.routable:
            print('We have %d routable IPs.' % (
                ivre.geoiputils.count_routable_ips()
            ))
            exit(0)
        parser.error("argument --output: invalid choice: '%s' "
                     "(only available with --country, --asnum, --region, "
                     "--city or --routable)" % args.output)
    if args.output in ['List', 'ListAll', 'ListCIDRs']:
        if args.country is not None:
            ivre.geoiputils.list_ips_by_country(
                args.country, listall=args.output == 'ListAll',
                listcidrs=args.output == 'ListCIDRs',
            )
            exit(0)
        if args.region is not None:
            ivre.geoiputils.list_ips_by_region(
                *args.region,
                listall=args.output == 'ListAll',
                listcidrs=args.output == 'ListCIDRs'
            )
            exit(0)
        if args.city is not None:
            ivre.geoiputils.list_ips_by_city(
                *args.city,
                listall=args.output == 'ListAll',
                listcidrs=args.output == 'ListCIDRs'
            )
            exit(0)
        if args.asnum is not None:
            ivre.geoiputils.list_ips_by_asnum(
                args.asnum, listall=args.output == 'ListAll',
                listcidrs=args.output == 'ListCIDRs',
            )
            exit(0)
        if args.routable:
            ivre.geoiputils.list_routable_ips(
                listall=args.output == 'ListAll',
                listcidrs=args.output == 'ListCIDRs',
            )
            exit(0)
        parser.error("argument --output: invalid choice: '%s' "
                     "(only available with --country, --region, --city, "
                     "--asnum or --routable)" % args.output)
    targets = ivre.target.target_from_args(args)
    if targets is None:
        parser.error('one argument of --country/--region/--city/--asnum/'
                     '--range/--network/--routable/--file/--test is required')
    if args.again is not None:
        accept_target_status = set(reduce(
            lambda x, y: x + y, [{
                'up': [STATUS_DONE_UP],
                'down': [STATUS_DONE_DOWN],
                'unknown': [STATUS_DONE_UNKNOWN],
                'all': [STATUS_DONE_UP, STATUS_DONE_DOWN,
                        STATUS_DONE_UNKNOWN]
            }[x] for x in args.again],
            [STATUS_NEW]))
    if args.zmap_prescan_port is not None:
        args.nmap_ping_types = ["PS%d" % args.zmap_prescan_port]
    elif args.nmap_prescan_ports is not None:
        args.nmap_ping_types = [
            "PS%s" % ",".join(str(p) for p in args.nmap_prescan_ports)
        ]
    options = ivre.nmapopt.build_nmap_options(template=args.nmap_template)
    if args.nmap_max_cpu is not None:
        NMAP_LIMITS[resource.RLIMIT_CPU] = (args.nmap_max_cpu,
                                            args.nmap_max_cpu)
    if args.nmap_max_heap_size is not None:
        NMAP_LIMITS[resource.RLIMIT_DATA] = (args.nmap_max_heap_size,
                                             args.nmap_max_heap_size)
    if args.nmap_max_stack_size is not None:
        NMAP_LIMITS[resource.RLIMIT_STACK] = (args.nmap_max_stack_size,
                                              args.nmap_max_stack_size)
    if args.output == 'XMLFork':
        pool = multiprocessing.Pool(processes=args.processes)
        if USE_PARTIAL:
            call_nmap_single = functools.partial(_call_nmap_single,
                                                 targets.infos[
                                                     'categories'][0],
                                                 options,
                                                 accept_target_status)
            for _ in pool.imap(call_nmap_single, targets, chunksize=1):
                pass
        else:
            for _ in pool.imap(
                    _call_nmap_single_tuple,
                    (
                        (targets.infos['categories'][0],
                         options,
                         accept_target_status,
                         target) for target in targets
                    ),
                    chunksize=1
            ):
                pass
        exit(0)
    elif args.output == 'ListAllRand':
        targiter = targets.__iter__()
        try:
            for target in targiter:
                print(ivre.utils.int2ip(target))
        except KeyboardInterrupt:
            print('Interrupted.\nUse "--state %s" to resume.' % (
                ' '.join(str(elt) for elt in targiter.getstate())
            ))
        except Exception:
            ivre.utils.LOGGER.critical('Exception', exc_info=True)
            print('Use "--state %s" to resume.' % (
                ' '.join(str(elt) for elt in targiter.getstate())
            ))
        exit(0)
    xmlprocess = {
        'XML': (XmlProcessWritefile,
                ['./scans/%s/' % targets.infos['categories'][0]], {}),
        'XMLFull': (XmlProcessWritefile,
                    ['./scans/%s/' % targets.infos['categories'][0]],
                    {'fulloutput': True}),
        'Test': (XmlProcessTest, [], {}),
    }[args.output]
    xmlprocess = xmlprocess[0](*xmlprocess[1], **xmlprocess[2])
    retval = call_nmap(options, xmlprocess, targets,
                       accept_target_status=accept_target_status)
    exit(retval)



#! /usr/bin/env python

# This file is part of IVRE.
# Copyright 2011 - 2018 Pierre LALET <pierre.lalet@cea.fr>
#
# IVRE is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IVRE is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with IVRE. If not, see <http://www.gnu.org/licenses/>.

"""Put selected results in views."""

from datetime import datetime

from ivre import utils
from ivre.xmlnmap import SCHEMA_VERSION, create_ssl_cert
from ivre.db import db


def _extract_passive_HTTP_CLIENT_HEADER_SERVER(rec):
    """Handle http client header about server."""
    return {'ports': [{
        'state_state': 'open',
        'state_reason': "passive",
        'port': rec['port'],
        'protocol': rec.get('protocol', 'tcp'),
        'service_name': 'http',
    }]}
    # TODO: (?) handle Host: header for DNS
    # FIXME: catches ip addresses as domain name.
    if 'source' in rec and rec['source'] == 'HOST':
        values = rec['value'].split(".")
        domains = [values.pop()]
        while values:
            domains.insert(0, values.pop() + "." + domains[0])
        return {'hostnames': [{'domains': domains,
                               'type': "?",
                               'name': domains[0]}]}


def _extract_passive_HTTP_SERVER_HEADER(rec):
    """Handle http server headers."""
    port = {
        'state_state': 'open',
        'state_reason': "passive",
        'port': rec['port'],
        'protocol': rec.get('protocol', 'tcp'),
        'service_name': 'http',
    }
    # TODO: handle other header values and merge them
    if rec.get('source') != 'SERVER':
        return {'ports': [port]}
    value = rec['value']
    script = {'id': 'http-server-header', 'output': value}
    port['scripts'] = [script]
    banner = (b"HTTP/1.1 200 OK\r\nServer: " + utils.nmap_decode_data(value) +
              b"\r\n\r\n")
    port.update(
        utils.match_nmap_svc_fp(output=banner,
                                proto=rec.get('protocol', 'tcp'),
                                probe="GetRequest")
    )
    return {'ports': [port]}


def _extract_passive_HTTP_CLIENT_HEADER(rec):
    """Handle http client headers."""
    # TODO: handle other header values
    if rec.get('source') != 'USER-AGENT':
        return {}
    return {'ports': [{
        'port': -1,
        'scripts': [{'id': 'http-user-agent',
                     'output': rec['value'],
                     'http-user-agent': [rec['value']]}],
    }]}


def _extract_passive_TCP_SERVER_BANNER(rec):
    """Handle banners from tcp servers."""
    value = rec['value']
    if rec['recontype'] == 'SSH_SERVER':
        value += "\r\n"
    port = {
        'state_state': 'open',
        'state_reason': "passive",
        'port': rec['port'],
        'protocol': rec.get('protocol', 'tcp'),
        'scripts': [{"id": "banner",
                     "output": value}],
    }
    port.update(rec.get('infos', {}))
    port.update(
        utils.match_nmap_svc_fp(output=utils.nmap_decode_data(value),
                                proto=rec.get('protocol', 'tcp'),
                                probe="NULL")
    )
    return {'ports': [port]}


_KEYS = {
    'ecdsa-sha2-nistp256': 'ECDSA',
}


def _extract_passive_SSH_SERVER_HOSTKEY(rec):
    """Handle SSH host keys."""
    # TODO: should (probably) be merged, sorted by date/time, keep one
    # entry per key type.
    #
    # (MAYBE) we should add a "lastseen" tag to every intel in view.
    value = utils.encode_b64(
        utils.nmap_decode_data(rec['value'])
    ).decode()
    fingerprint = rec['infos']['md5']
    key = {'type': rec['infos']['algo'],
           'key': value,
           'fingerprint': fingerprint}
    if 'bits' in rec['infos']:  # FIXME
        key['bits'] = rec['infos']['bits']
    fingerprint = utils.decode_hex(fingerprint)
    script = {
        'id': 'ssh-hostkey', 'ssh-hostkey': [key],
        'output': '\n  %s %s (%s)\n%s %s' % (
            key.get('bits', '-'),  # FIXME
            ':'.join('%02x' % (
                ord(i) if isinstance(i, (bytes, str)) else i
            ) for i in fingerprint),
            _KEYS.get(
                key['type'],
                (key['type'][4:] if key['type'][:4] == 'ssh-'
                 else key['type']).upper()
            ),
            key['type'],
            value
        ),
        'key': key
    }
    return {'ports': [{
        'state_state': 'open',
        'state_reason': "passive",
        'port': rec['port'],
        'protocol': rec.get('protocol', 'tcp'),
        'service_name': 'ssh',
        'scripts': [script],
    }]}


def _extract_passive_SSL_SERVER(rec):
    """Handle ssl server headers."""
    source = rec.get('source')
    if source == 'cert':
        return _extract_passive_SSL_SERVER_cert(rec)
    if source.startswith('ja3-'):
        return _extract_passive_SSL_SERVER_ja3(rec)
    return {}


def _extract_passive_SSL_SERVER_cert(rec):
    script = {"id": "ssl-cert"}
    port = {
        'state_state': 'open',
        'state_reason': "passive",
        'port': rec['port'],
        'protocol': rec.get('protocol', 'tcp'),
    }
    output, info = create_ssl_cert(rec['value'], b64encoded=False)
    if info:
        script['output'] = "\n".join(output)
        script['ssl-cert'] = info
        port['scripts'] = [script]
    return {'ports': [port]}


def _extract_passive_SSL_SERVER_ja3(rec):
    script = {"id": "ssl-ja3-server"}
    port = {
        'state_state': 'open',
        'state_reason': 'passive',
        'port': rec['port'],
        'protocol': rec.get('protocol', 'tcp'),
    }
    script['output'] = rec['value'] + ' - ' + rec['source'][4:]
    info = {
        'raw': rec['infos']['raw'],
        'sha256': rec['infos']['sha256'],
        'sha1': rec['infos']['sha1'],
        'md5': rec['value'],
        'client': {
            'raw': rec['infos']['client']['raw'],
            'sha256': rec['infos']['client']['sha256'],
            'sha1': rec['infos']['client']['sha1'],
            'md5': rec['source'][4:]
        }
    }
    script['ssl-ja3-server'] = [info]
    port['scripts'] = [script]
    return {'ports': [port]}


def _extract_passive_DNS_ANSWER(rec):
    """Handle dns server headers."""
    name = rec['value']
    domains = rec['infos']['domain']
    return {'hostnames': [{'domains': domains,
                           'type': rec['source'].split('-', 1)[0],
                           'name': name}]}


def _extract_passive_SSL_CLIENT(rec):
    """Handle SSL client ja3 extraction."""
    script = {"id": "ssl-ja3-client"}
    script['output'] = rec['value']
    script['ssl-ja3-client'] = [{
        'raw': rec['infos']['raw'],
        'sha256': rec['infos']['sha256'],
        'sha1': rec['infos']['sha1'],
        'md5': rec['value']
    }]

    port = {
        'port': -1,
        'scripts': [script]
    }

    return {'ports': [port]}


_EXTRACTORS = {
    # 'HTTP_CLIENT_HEADER_SERVER': _extract_passive_HTTP_CLIENT_HEADER_SERVER,
    'HTTP_CLIENT_HEADER': _extract_passive_HTTP_CLIENT_HEADER,
    'HTTP_SERVER_HEADER': _extract_passive_HTTP_SERVER_HEADER,
    'SSL_SERVER': _extract_passive_SSL_SERVER,
    'SSL_CLIENT': _extract_passive_SSL_CLIENT,
    # FIXME: see db/prostgres while hostnames are not merged, it is useless
    # to add DNS answers. It creates empty results.
    'DNS_ANSWER': _extract_passive_DNS_ANSWER,
    'SSH_SERVER': _extract_passive_TCP_SERVER_BANNER,
    'SSH_SERVER_HOSTKEY': _extract_passive_SSH_SERVER_HOSTKEY,
    'TCP_SERVER_BANNER': _extract_passive_TCP_SERVER_BANNER,
}


def passive_record_to_view(rec):
    """Return a passive entry in the View format.

    Note that this entry is likely to have no sense in itself. This
    function is intended to be used to format results for the merge
    function.

    """
    rec = dict(rec)
    if not rec.get('addr'):
        return
    outrec = {
        'addr': rec["addr"],
        'state_reason': 'passive',
        'schema_version': SCHEMA_VERSION,
    }
    # a DNS_ANSWER record is not enough to mark a host as up
    if rec['recontype'] != 'DNS_ANSWER':
        outrec['state'] = 'up'
    sensor = rec.get('sensor')
    if sensor:
        outrec['source'] = [sensor]
    try:
        outrec['starttime'] = datetime.fromtimestamp(rec["firstseen"])
        outrec['endtime'] = datetime.fromtimestamp(rec["lastseen"])
    except TypeError:
        outrec['starttime'] = rec['firstseen']
        outrec['endtime'] = rec['lastseen']
    function = _EXTRACTORS.get(rec['recontype'], lambda _: {})
    if isinstance(function, dict):
        function = function.get(rec['source'], lambda _: {})
    outrec.update(function(rec))
    openports = outrec['openports'] = {'count': 0}
    for port in outrec.get('ports', []):
        if port.get('state_state') != 'open':
            continue
        openports['count'] += 1
        protoopenports = openports.setdefault(port['protocol'],
                                              {'count': 0, 'ports': []})
        protoopenports['count'] += 1
        protoopenports['ports'].append(port['port'])
    return outrec


def passive_to_view(flt):
    """Generates passive entries in the View format.

    Note that this entry is likely to have no sense in itself. This
    function is intended to be used to format results for the merge
    function.

    """
    for rec in db.passive.get(flt, sort=[("addr", 1)]):
        outrec = passive_record_to_view(rec)
        if outrec is not None:
            yield outrec


def from_passive(flt):
    """Iterator over passive results, by address."""
    records = passive_to_view(flt)
    cur_addr = None
    cur_rec = None
    for rec in records:
        if cur_addr is None:
            cur_addr = rec['addr']
            cur_rec = rec
        elif cur_addr != rec['addr']:
            # TODO: add_addr_info should be optional
            cur_rec['infos'] = {}
            for func in [db.data.country_byip,
                         db.data.as_byip,
                         db.data.location_byip]:
                cur_rec['infos'].update(func(cur_addr) or {})
            yield cur_rec
            cur_rec = rec
            cur_addr = rec['addr']
        else:
            cur_rec = db.view.merge_host_docs(cur_rec, rec)
    if cur_rec is not None:
        yield cur_rec


def nmap_record_to_view(rec):
    """Convert an nmap result in view.

    """
    if 'scanid' in rec:
        del rec['scanid']
    if 'source' in rec:
        if not rec['source']:
            rec['source'] = []
        elif not isinstance(rec['source'], list):
            rec['source'] = [rec['source']]
    return rec


def from_nmap(flt):
    """Return an Nmap entry in the View format."""
    cur_addr = None
    cur_rec = None
    result = None
    for rec in db.nmap.get(flt, sort=[("addr", 1)]):
        if 'addr' not in rec:
            continue
        rec = nmap_record_to_view(rec)
        if cur_addr is None:
            cur_addr = rec['addr']
            cur_rec = rec
            continue
        if cur_addr != rec['addr']:
            result = cur_rec
            cur_rec = rec
            cur_addr = rec['addr']
            yield result
        else:
            cur_rec = db.view.merge_host_docs(cur_rec, rec)
            continue
    if cur_rec is not None:
        yield cur_rec


def to_view(itrs):
    """Takes a list of iterators over view-formated results, and returns an
    iterator over merged results, sorted by ip.

    """

    def next_record(rec, updt):
        if rec is None:
            return updt
        return db.view.merge_host_docs(rec, updt)
    next_recs = []
    # We cannot use a `for itr in itrs` loop here because itrs is
    # modified in the loop.
    i = 0
    while i < len(itrs):
        try:
            next_recs.append(next(itrs[i]))
        except StopIteration:
            # We need to remove the corresponding iterator from itrs,
            # which happens to be the n-th where n is the current
            # length of next_recs.
            del itrs[len(next_recs)]  # Do not increment i here
        else:
            i += 1
    next_addrs = [rec['addr'] for rec in next_recs]
    cur_rec = None
    try:
        cur_addr = min(next_addrs, key=utils.ip2int)
    except ValueError:
        # next_addrs is empty
        cur_addr = None
    while next_recs:
        # We cannot use a `for i in range(len(itrs))` loop because
        # itrs is modified in the loop.
        i = 0
        while i < len(itrs):
            if next_addrs[i] == cur_addr:
                cur_rec = next_record(cur_rec, next_recs[i])
                try:
                    next_recs[i] = next(itrs[i])
                except StopIteration:
                    del next_addrs[i]
                    del next_recs[i]
                    del itrs[i]
                    continue  # Do not increment i here
                next_addrs[i] = next_recs[i]['addr']
            i += 1
        if next_addrs and cur_addr not in next_addrs:
            yield cur_rec
            cur_rec = None
            cur_addr = min(next_addrs)
    if cur_rec is not None:
        yield cur_rec



#! /usr/bin/env python

# This file is part of IVRE.
# Copyright 2011 - 2019 Pierre LALET <pierre.lalet@cea.fr>
#
# IVRE is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IVRE is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with IVRE. If not, see <http://www.gnu.org/licenses/>.

"""This sub-module contains functions to interact with another IVRE
instance via an HTTP server.

"""


import json
try:
    from urllib.request import URLopener
except ImportError:
    from urllib import URLopener


from ivre.db import DB, DBActive, DBNmap, DBView


class HttpDB(DB):

    flt_empty = ""

    def __init__(self, url):
        super(HttpDB, self).__init__()
        self.baseurl = url._replace(fragment="").geturl()
        self.db = urlop = URLopener()
        for hdr, val in (
            tuple(x.split("=", 1)) if "=" in x else (x, "")
            for x in url.fragment.split("&")
            if x
        ):
            urlop.addheader(hdr, val)

    def get(self, spec, limit=None, skip=None, sort=None, fields=None):
        url = '%s/%s?q=%sskip:' % (self.baseurl, self.route,
                                   ('%s%%20' % spec) if spec else '')
        if skip is None:
            skip = 0
        while True:
            cururl = '%s%d' % (url, skip)
            if limit is not None:
                cururl += '%%20limit:%d' % limit
            req = self.db.open(cururl)
            data = json.loads(req.read().decode())
            if not data:
                break
            if limit is None:
                for rec in data:
                    yield rec
            else:
                for rec in data:
                    yield rec
                    limit -= 1
                    if limit == 0:
                        break
                if limit == 0:
                    break
            skip += len(data)

    def count(self, spec, **kargs):
        url = '%s/%s/count' % (self.baseurl, self.route)
        if spec:
            url += '?q=%s' % spec
        req = self.db.open(url)
        return int(req.read().rstrip(b'\n'))

    @staticmethod
    def flt_and(*args):
        return '%20'.join(arg for arg in args if arg)


class HttpDBActive(HttpDB, DBActive):

    @staticmethod
    def searchhost(addr, neg=False):
        return '%s%s' % ('!' if neg else '', addr)


class HttpDBNmap(HttpDBActive, DBNmap):

    route = 'scans'


class HttpDBView(HttpDBActive, DBView):

    route = 'view'



#! /usr/bin/env python

# This file is part of IVRE.
# Copyright 2011 - 2018 Pierre LALET <pierre.lalet@cea.fr>
#
# IVRE is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IVRE is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with IVRE. If not, see <http://www.gnu.org/licenses/>.

"""This sub-module contains functions to interact with the MaxMind
files.

"""


import codecs
from functools import reduce
import os
import struct


from builtins import object, range
from future.utils import viewitems


from ivre import config, utils
from ivre.db import DBData


class MaxMindFileIter(object):

    """Iterator for MaxMindFile"""

    def __init__(self, base):
        self.base = base
        self.current = []
        self.nextval = 0

    def __iter__(self):
        return self

    def __next__(self):
        if self.nextval is None:
            raise StopIteration()
        node_no = 0
        for i in range(96 if self.base.ip_version == 4 else 0, 128):
            try:
                flag = self.current[i]
            except IndexError:
                flag = self.nextval
                self.current.append(self.nextval)
                self.nextval = 0
            next_node_no = self.base.read_record(node_no, flag)
            if next_node_no == 0:
                raise Exception('Invalid file format')
            elif next_node_no >= self.base.node_count:
                pos = (next_node_no - self.base.node_count -
                       self.base.DATA_SECTION_SEPARATOR_SIZE)
                curvalinf = int(''.join(str(p) for p in self.current) +
                                '0' * (128 - len(self.current)), 2)
                curvalsup = int(''.join(str(p) for p in self.current) +
                                '1' * (128 - len(self.current)), 2)
                try:
                    while self.current.pop():
                        pass
                except IndexError:
                    self.nextval = None
                else:
                    self.current.append(1)
                return (curvalinf, curvalsup,
                        self.base.decode(pos, self.base.data_section_start)[1])
            node_no = next_node_no


class EmptyMaxMindFile(object):

    """Stub to replace MaxMind databases parsers. Used when a file is
missing to emit a warning message and return empty results.

    """

    def __init__(self, _):
        utils.LOGGER.warning("Cannot find Maxmind database files")

    def lookup(self, _):
        return {}


class MaxMindFile(object):

    """Parser for MaxMind databases.

    Code copied and adapted from
    https://github.com/yhirose/maxminddb/blob/master/lib/maxminddb.rb

    """

    METADATA_BEGIN_MARKER = b'\xab\xcd\xefMaxMind.com'
    DATA_SECTION_SEPARATOR_SIZE = 16
    SIZE_BASE_VALUES = [0, 29, 285, 65821]
    POINTER_BASE_VALUES = [0, 0, 2048, 526336]

    def __init__(self, path):
        self.path = path
        self._data = None
        pos = (self.data.rindex(self.METADATA_BEGIN_MARKER) +
               len(self.METADATA_BEGIN_MARKER))
        metadata = self.metadata = self.decode(pos, 0)[1]
        self.ip_version = metadata['ip_version']
        self.node_count = metadata['node_count']
        self.node_byte_size = metadata['record_size'] * 2 // 8
        self.search_tree_size = self.node_count * self.node_byte_size
        self.data_section_start = (self.search_tree_size +
                                   self.DATA_SECTION_SEPARATOR_SIZE)

    @property
    def data(self):
        if self._data is None:
            with open(self.path, 'rb') as fdesc:
                self._data = fdesc.read()
        return self._data

    def read_byte(self, pos):
        return ord(self.data[pos:pos + 1])

    def read_value(self, pos, size):
        return reduce(
            lambda x, y: (x << 8) + y,
            struct.unpack(
                '%dB' % size,
                self.data[pos:pos + size],
            ),
            0,
        )

    def decode(self, pos, base_pos):
        ctrl = ord(self.data[pos + base_pos:pos + base_pos + 1])
        pos += 1
        type_ = ctrl >> 5
        if type_ == 1:
            # pointer
            size = ((ctrl >> 3) & 0x3) + 1
            val1 = ctrl & 0x7
            val2 = self.read_value(pos + base_pos, size)
            pointer = ((val1 << (8 * size)) + val2 +
                       self.POINTER_BASE_VALUES[size])
            return pos + size, self.decode(pointer, base_pos)[1]
        if type_ == 0:
            # extended type
            type_ = 7 + self.read_byte(pos + base_pos)
            pos += 1
        size = ctrl & 0x1f
        if size >= 29:
            byte_size = size - 29 + 1
            val = self.read_value(pos + base_pos, byte_size)
            pos += byte_size
            size = val + self.SIZE_BASE_VALUES[byte_size]
        if type_ == 2:
            # utf8
            val = self.data[
                pos + base_pos:pos + base_pos + size
            ].decode('utf-8')
            pos += size
        elif type_ in [3, 15]:
            # double
            # float
            val = struct.unpack(
                {3: '>d', 15: '>f'}[type_],
                self.data[pos + base_pos:pos + base_pos + size],
            )[0]
            pos += size
        elif type_ == 4:
            # bytes
            val = self.data[pos + base_pos:pos + base_pos + size]
            pos += size
        elif type_ in [5, 6, 9, 10]:
            # unsigned 16-bit int
            # unsigned 32-bit int
            # unsigned 64-bit int
            # unsigned 128-bit int
            val = self.read_value(pos + base_pos, size)
            pos += size
        elif type_ == 7:
            # map
            val = {}
            for _ in range(size):
                pos, k = self.decode(pos, base_pos)
                pos, v = self.decode(pos, base_pos)
                val[k] = v
        elif type_ == 8:
            # signed 32-bit int
            v1 = struct.unpack(
                '>i', self.data[pos + base_pos:pos + base_pos + size]
            )[0]
            bits = size * 8
            val = (v1 & ~(1 << bits)) - (v1 & (1 << bits))
            pos += size
        elif type_ == 11:
            # array
            val = []
            for _ in range(size):
                pos, v = self.decode(pos, base_pos)
                val.append(v)
        elif type_ == 12:
            # data cache container
            raise Exception('TODO type == 12 (data cache container)')
        elif type_ == 13:
            # end marker
            val = None
        elif type_ == 14:
            # boolean
            val = bool(size)
        else:
            # unknown
            raise Exception('TODO type == %d (unknown)' % type_)
        return pos, val

    def read_record(self, node_no, flag):
        rec_byte_size = self.node_byte_size // 2
        pos = self.node_byte_size * node_no
        middle = (self.read_byte(pos + rec_byte_size)
                  if self.node_byte_size % 2 else 0)
        if flag == 0:  # left
            val = self.read_value(pos, rec_byte_size)
            val += ((middle & 0xf0) << 20) if middle else 0
        else:  # right
            val = self.read_value(pos + self.node_byte_size - rec_byte_size,
                                  rec_byte_size)
            val += ((middle & 0xf) << 24) if middle else 0
        return val

    def __repr__(self):
        return '<%s from %s>' % (self.__class__.__name__, self.path)

    def lookup(self, ip):
        node_no = 0
        addr = utils.force_ip2int(ip)
        for i in range(96 if self.ip_version == 4 else 0, 128):
            flag = (addr >> (127 - i)) & 1
            next_node_no = self.read_record(node_no, flag)
            if next_node_no == 0:
                raise Exception('Invalid file format')
            elif next_node_no >= self.node_count:
                pos = (next_node_no - self.node_count -
                       self.DATA_SECTION_SEPARATOR_SIZE)
                return self.decode(pos, self.data_section_start)[1]
            else:
                node_no = next_node_no
        raise Exception('Invalid file format')

    def __iter__(self):
        return MaxMindFileIter(self)

    @staticmethod
    def _get_fields(rec, fields):
        for field in fields:
            val = rec
            for subfield in field.split('->'):
                try:
                    val = val[subfield]
                except TypeError:
                    try:
                        subfield = int(subfield)
                    except ValueError:
                        val = None
                        break
                    try:
                        val = val[subfield]
                    except IndexError:
                        val = None
                        break
                except KeyError:
                    val = None
                    break
            yield val

    def _get_ranges(self, fields):
        gen = iter(self)
        start, stop, rec = next(gen)
        rec = tuple(self._get_fields(rec, fields))
        for n_start, n_stop, n_rec in gen:
            n_rec = tuple(self._get_fields(n_rec, fields))
            if n_start <= stop + 1 and n_rec == rec:
                stop = n_stop
            else:
                yield (start, stop) + rec
                start, stop, rec = n_start, n_stop, n_rec
        yield (start, stop) + rec

    def get_ranges(self, fields, cond=None):
        for rec in self._get_ranges(fields):
            if cond is None or cond(rec):
                yield rec


class MaxMindDBData(DBData):

    LANG = "en"

    AS_KEYS = {
        "autonomous_system_number": "as_num",
        "autonomous_system_organization": "as_name",
    }

    @property
    def db_asn(self):
        try:
            return self._db_asn
        except AttributeError:
            self._db_asn = EmptyMaxMindFile("ASN")
            return self._db_asn

    @property
    def db_city(self):
        try:
            return self._db_city
        except AttributeError:
            self._db_city = EmptyMaxMindFile("City")
            return self._db_city

    @property
    def db_country(self):
        try:
            return self._db_country
        except AttributeError:
            self._db_country = EmptyMaxMindFile("Country")
            return self._db_country

    def __init__(self, url):
        self.basepath = url.path
        self.reload_files()

    def reload_files(self):
        for fname in os.listdir(self.basepath):
            if fname.endswith('.mmdb'):
                subdb = MaxMindFile(os.path.join(self.basepath, fname))
                name = subdb.metadata['database_type'].lower()
                if name.startswith('geolite2-'):
                    name = name[9:]
                setattr(self, "_db_%s" % name, subdb)

    def as_byip(self, addr):
        return dict(
            (self.AS_KEYS.get(key, key), value)
            for key, value in viewitems(self.db_asn.lookup(addr))
        )

    def location_byip(self, addr):
        raw = self.db_city.lookup(addr)
        result = {}
        sub = raw.get('subdivisions')
        if sub:
            result['region_code'] = tuple(v.get('iso_code') for v in sub)
            result['region_name'] = tuple(v.get('names', {}).get(self.LANG)
                                          for v in sub)
        sub = raw.get('continent')
        if sub:
            value = sub.get('code')
            if value:
                result['continent_code'] = value
            value = sub.get('names', {}).get(self.LANG)
            if value:
                result['continent_name'] = value
        sub = raw.get('country')
        if sub:
            value = sub.get('iso_code')
            if value:
                result['country_code'] = value
            value = sub.get('names', {}).get(self.LANG)
            if value:
                result['country_name'] = value
        sub = raw.get('registered_country')
        if sub:
            value = sub.get('iso_code')
            if value:
                result['registered_country_code'] = value
            value = sub.get('names', {}).get(self.LANG)
            if value:
                result['registered_country_name'] = value
        value = raw.get('city', {}).get('names', {}).get(self.LANG)
        if value:
            result['city'] = value
        value = raw.get('postal', {}).get('code')
        if value:
            result['postal_code'] = value
        sub = raw.get('location')
        if sub:
            try:
                result['coordinates'] = (sub['latitude'], sub['longitude'])
            except KeyError:
                pass
            value = sub.get('accuracy_radius')
            result['coordinates_accuracy_radius'] = value
        if result:
            return result

    def country_byip(self, addr):
        result = {}
        raw = self.db_country.lookup(addr)
        sub = raw.get('country')
        if sub:
            value = sub.get('iso_code')
            if value:
                result['country_code'] = value
            value = sub.get('names', {}).get(self.LANG)
            if value:
                result['country_name'] = value
        return result

    def dump_as_ranges(self, fdesc):
        for data in self.db_asn.get_ranges(
                ["autonomous_system_number"],
                cond=lambda line: line[2] is not None,
        ):
            if data[0] > 0xffffffff:  # only IPv4
                break
            fdesc.write('%d,%d,%d\n' % data)

    def dump_country_ranges(self, fdesc):
        for data in self.db_country.get_ranges(
                ["country->iso_code"],
                cond=lambda line: line[2] is not None,
        ):
            if data[0] > 0xffffffff:  # only IPv4
                break
            fdesc.write('%d,%d,%s\n' % data)

    def dump_city_ranges(self, fdesc):
        for data in self.db_city.get_ranges(
                ["country->iso_code", "subdivisions->0->iso_code",
                 "city->names->%s" % config.GEOIP_LANG, "city->geoname_id"],
                cond=lambda line: (line[2] is not None and
                                   (line[3] is not None or
                                    line[4] is not None)),
        ):
            if data[0] > 0xffffffff:  # only IPv4
                break
            fdesc.write('%d,%d,%s,%s,%s,%s\n' % (
                data[:4] +
                (utils.encode_b64((data[4] or
                                   "").encode('utf-8')).decode('utf-8'),) +
                data[5:]
            ))

    def build_dumps(self, force=False):
        for attr, func in [
                ("db_asn", self.dump_as_ranges),
                ("db_country", self.dump_country_ranges),
                ("db_city", self.dump_city_ranges),
        ]:
            try:
                subdb = getattr(self, attr)
            except AttributeError:
                continue
            if not subdb.path.endswith('.mmdb'):
                continue
            csv_file = subdb.path[:-4] + 'dump-IPv4.csv'
            if not force:
                mmdb_mtime = os.path.getmtime(subdb.path)
                try:
                    csv_mtime = os.path.getmtime(csv_file)
                except OSError:
                    pass
                else:
                    if csv_mtime > mmdb_mtime:
                        utils.LOGGER.info('Skipping %r since %r is newer',
                                          os.path.basename(subdb.path),
                                          os.path.basename(csv_file))
                        continue
            utils.LOGGER.info('Dumping %r to %r', os.path.basename(subdb.path),
                              os.path.basename(csv_file))
            with codecs.open(csv_file, mode="w", encoding='utf-8') as fdesc:
                func(fdesc)



#! /usr/bin/env python

# This file is part of IVRE.
# Copyright 2011 - 2019 Pierre LALET <pierre.lalet@cea.fr>
#
# IVRE is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# IVRE is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with IVRE. If not, see <http://www.gnu.org/licenses/>.


from __future__ import print_function

try:
    import argparse
except ImportError:
    import optparse
    USING_ARGPARSE = False
else:
    USING_ARGPARSE = True
import sys
try:
    reload(sys)
except NameError:
    pass
else:
    sys.setdefaultencoding('utf-8')


from ivre import utils


def main():
    if USING_ARGPARSE:
        parser = argparse.ArgumentParser(
            description='Tool for ip addresses manipulation.',
        )
    else:
        parser = optparse.OptionParser(
            description='Tool for ip addresses manipulation.',)
        parser.parse_args_orig = parser.parse_args

        def my_parse_args():
            res = parser.parse_args_orig()
            res[0].ensure_value('ips', res[1])
            return res[0]
        parser.parse_args = my_parse_args
        parser.add_argument = parser.add_option

    if USING_ARGPARSE:
        parser.add_argument('ips', nargs='*',
                            help='Display results for specified IP addresses'
                            ' or ranges.')
    args = parser.parse_args()

    for a in args.ips:
        if '/' in a:
            a = utils.net2range(a)
            print("%s-%s" % (a[0], a[1]))
        elif '-' in a:
            a = a.split('-', 1)
            if a[0].isdigit():
                a[0] = int(a[0])
            if a[1].isdigit():
                a[1] = int(a[1])
            for n in utils.range2nets((a[0], a[1])):
                print(n)
        else:
            if a.isdigit():
                a = utils.force_int2ip(int(a))
            else:
                a = utils.force_ip2int(a)
            print(a)
